
import abc
import asyncio
import copy
import functools
import logging
import sys
import time

from tornado import queues

from qtoggleserver import persist
from qtoggleserver import utils

from qtoggleserver.conf import settings
from qtoggleserver.core import events as core_events
from qtoggleserver.core import expressions as core_expressions
from qtoggleserver.core import main
from qtoggleserver.core import sessions as core_sessions
from qtoggleserver.core import sequences as core_sequences
from qtoggleserver.utils import json as json_utils


TYPE_BOOLEAN = 'boolean'
TYPE_NUMBER = 'number'

CHANGE_REASON_NATIVE = 'N'
CHANGE_REASON_API = 'A'
CHANGE_REASON_SEQUENCE = 'S'
CHANGE_REASON_EXPRESSION = 'E'

logger = logging.getLogger(__name__)

_ports = {}  # indexed by id


STANDARD_ATTRDEFS = {
    'id': {
        'type': 'string'
    },
    'display_name': {
        'type': 'string',
        'modifiable': True,
        'max': 64
    },
    'type': {
        'type': 'string',
        'choices': [
            {
                'value': 'boolean',
                'display_name': 'Boolean'
            },
            {
                'value': 'number',
                'display_name': 'Number'
            }
        ]
    },
    'unit': {
        'type': 'string',
        'modifiable': True,
        'max': 16
    },
    'writable': {
        'type': 'boolean'
    },
    'enabled': {
        'type': 'boolean',
        'modifiable': True
    },

    'min': {
        'type': 'number',
        'optional': True
    },
    'max': {
        'type': 'number',
        'optional': True
    },
    'integer': {
        'type': 'boolean',
        'optional': True
    },
    'step': {
        'type': 'number',
        'optional': True
    },
    'choices': {  # TODO data type uncertain
        'type': '[]',
        'optional': True
    },

    'tag': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 64,
    },
    'expression': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 1024,
    },
    'transform_read': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 1024,
    },
    'transform_write': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 1024,
    },
    'persisted': {
        'type': 'boolean',
        'optional': True,
        'modifiable': True
    },
    'virtual': {
        'type': 'boolean',
        'optional': True
    },
    'online': {
        'type': 'boolean',
        'optional': True
    }
}


class PortError(Exception):
    pass


class InvalidAttributeValue(PortError):
    def __init__(self, attr):
        self.attr = attr

        super().__init__(attr)


class PortTimeout(PortError):
    pass


class BasePort(utils.LoggableMixin, metaclass=abc.ABCMeta):
    PERSIST_COLLECTION = 'ports'

    TYPE = TYPE_BOOLEAN
    DISPLAY_NAME = ''
    UNIT = ''
    WRITABLE = False
    CHOICES = None

    WRITE_VALUE_PAUSE = 0

    STANDARD_ATTRDEFS = STANDARD_ATTRDEFS

    ADDITIONAL_ATTRDEFS = {}
    '''
    ADDITIONAL_ATTRDEFS = {
        'attr1': {
            'display_name': 'Some Attribute Display Name',
            'description': 'Some attribute description',
            'type': 'number',
            'modifiable': True,
            'pattern': '^.*$',
            'unit': 'seconds',
            'min': 1,
            'max': 100,
            'integer': True,
            'step': 5,
            'choices': [
                {'value': 2, 'display_name': 'Two'},
                {'value': 4, 'display_name': 'Four'}
            ]
        },
        ...
    }'''

    def __init__(self, port_id):
        utils.LoggableMixin.__init__(self, port_id, logger)

        self._id = port_id
        self._enabled = False
        self._display_name = None
        self._unit = None

        self._sequence = None
        self._expression = None
        self._transform_read = None
        self._transform_write = None

        # attributes cache is used to prevent computing
        # an attribute value more than once per core iteration
        self._attrs_cache = {}

        # cache attribute definitions so that the ATTRDEFS property
        # doesn't need to gather all of them with each access
        self._attrdefs_cache = None

        self._modifiable_attrs = None
        self._non_modifiable_attrs = None

        self._schema = None
        self._value_schema = None

        self._value = None
        self._write_value_queue = queues.Queue()
        self._write_value_task = asyncio.create_task(self._write_value_loop())
        self._asap_value = None
        self._change_reason = CHANGE_REASON_NATIVE

        self._loaded = False

    def __str__(self):
        return 'port {}'.format(self._id)

    def __repr__(self):
        return 'Port({})'.format(self._id)

    def initialize(self):
        pass

    def _get_attrdefs(self):
        if self._attrdefs_cache is None:
            self._attrdefs_cache = dict(self.STANDARD_ATTRDEFS, **self.ADDITIONAL_ATTRDEFS)

        return self._attrdefs_cache

    ATTRDEFS = property(_get_attrdefs)

    def invalidate_attrdefs(self):
        self._attrs_cache = {}
        self._attrdefs_cache = None
        self._modifiable_attrs = None
        self._non_modifiable_attrs = None
        self._schema = None

    def get_non_modifiable_attrs(self):
        if self._non_modifiable_attrs is None:
            self._non_modifiable_attrs = set([n for (n, v) in self.ATTRDEFS.items() if not v.get('modifiable')])

        return self._non_modifiable_attrs

    def get_modifiable_attrs(self):
        if self._modifiable_attrs is None:
            self._modifiable_attrs = set([n for (n, v) in self.ATTRDEFS.items() if v.get('modifiable')])

        return self._modifiable_attrs

    async def get_attrs(self):
        d = {}

        for name in self.ATTRDEFS.keys():
            v = await self.get_attr(name)
            if v is None:
                continue

            d[name] = v

        return d

    def invalidate_attrs(self):
        self._attrs_cache = {}

    def invalidate_attr(self, name):
        self._attrs_cache.pop(name, None)

    async def get_attr(self, name):
        value = self._attrs_cache.get(name)
        if value is not None:
            return value

        method = getattr(self, 'attr_get_' + name, getattr(self, 'attr_is_' + name, None))
        if method:
            value = await method()
            if value is not None:
                self._attrs_cache[name] = value
                return value

        value = getattr(self, '_' + name, None)
        if value is not None:
            self._attrs_cache[name] = value
            return value

        method = getattr(self, 'attr_get_default_' + name, getattr(self, 'attr_is_default_' + name, None))
        if method:
            value = await method()
            if value is not None:
                self._attrs_cache[name] = value
                return value

        value = getattr(self, name.upper(), None)
        if value is not None:
            self._attrs_cache[name] = value
            return value

        return None  # unsupported attribute

    async def set_attr(self, name, value):
        old_value = await self.get_attr(name)
        if old_value is None:
            return  # refuse to set an unsupported attribute

        method = getattr(self, 'attr_set_' + name, None)
        if method:
            try:
                await method(value)

            except Exception:
                self.error('failed to set attribute %s = %s', name, json_utils.dumps(value), exc_info=True)

                raise

        elif hasattr(self, '_' + name):
            setattr(self, '_' + name, value)

        if self.is_loaded():
            await main.update()

        # new attributes might have been added or removed after setting an attribute;
        # therefore new definitions might have appeared or disappeared
        self.invalidate_attrdefs()

        self._attrs_cache[name] = value
        if old_value != value:
            await self.handle_attr_change(name, value)

        if not self.is_loaded():
            return

        # skip an IO loop iteration, allowing setting multiple attributes before triggering a port-update
        await asyncio.sleep(0)
        self.trigger_update()

    async def handle_attr_change(self, name, value):
        method_name = 'handle_{}_change'.format(name)
        method = getattr(self, method_name, None)
        if method:
            try:
                await method(value)

            except Exception as e:
                self.error('%s failed: %s', method_name, e, exc_info=True)

    def get_display_name(self):
        return self._display_name or self._id

    def get_id(self):
        return self._id

    def map_id(self, new_id):
        self._id = new_id
        self.debug('mapped to %s', new_id)

    async def get_type(self):
        return await self.get_attr('type')

    async def is_writable(self):
        return await self.get_attr('writable')

    async def is_persisted(self):
        return await self.get_attr('persisted')

    def is_enabled(self):
        return self._enabled

    async def enable(self):
        if self._enabled:
            return

        self.debug('enabling')
        self._enabled = True
        self.invalidate_attr('enabled')

        try:
            await self.handle_enable()

        except Exception:
            self.error('failed to enable')
            self._enabled = False

            raise

    async def disable(self):
        if not self._enabled:
            return

        # cancel sequence
        if self._sequence:
            self.debug('canceling current sequence')
            self._sequence.cancel()
            self._sequence = None

        self.debug('disabling')
        self._enabled = False
        self.invalidate_attr('enabled')

        try:
            await self.handle_disable()

        except Exception:
            self.error('failed to disable')
            self._enabled = True

            raise

    async def handle_enable(self):
        pass

    async def handle_disable(self):
        pass

    async def attr_set_enabled(self, value):
        if value:
            await self.enable()

        else:
            await self.disable()

    async def get_expression(self):
        if not await self.is_writable():
            return None

        return self._expression

    async def attr_get_expression(self):
        if not await self.is_writable():
            return None

        if self._expression:
            return str(self._expression)

        else:
            return ''

    async def attr_set_expression(self, sexpression):
        if not await self.is_writable():
            self.debug('refusing to set expression to non-writable port')
            return False

        if self._sequence:
            self.debug('canceling current sequence')
            self._sequence.cancel()

        if not sexpression:
            self._expression = None
            return

        try:
            self.debug('parsing expression "%s"', sexpression)
            expression = core_expressions.parse(self.get_id(), sexpression)

            self.debug('checking for expression circular dependencies')
            await core_expressions.check_loops(self, expression)

        except core_expressions.ExpressionError as e:
            self.error('failed to set expression "%s": %s', sexpression, e)

            raise InvalidAttributeValue('expression') from e

        self.debug('setting expression "%s"', expression)
        self._expression = expression

        main.force_eval_expressions()

    async def attr_get_transform_read(self):
        if self._transform_read:
            return str(self._transform_read)

        else:
            return ''

    async def attr_set_transform_read(self, stransform_read):
        if not stransform_read:
            self._transform_read = None
            return

        try:
            self.debug('parsing expression "%s"', stransform_read)
            transform_read = core_expressions.parse(self.get_id(), stransform_read)

            deps = transform_read.get_deps()
            for dep in deps:
                if not dep.startswith('$'):
                    continue

                if dep != '${}'.format(self._id):
                    raise core_expressions.ExpressionError('transform expression depends other ports')

            self.debug('setting read transform "%s"', transform_read)
            self._transform_read = transform_read

        except core_expressions.ExpressionError as e:
            self.error('failed to set transform read expression "%s": %s', stransform_read, e)

            raise InvalidAttributeValue('transform_read') from e

    async def attr_get_transform_write(self):
        if not await self.is_writable():
            return None  # only writable ports have transform_write attributes

        if self._transform_write:
            return str(self._transform_write)

        else:
            return ''

    async def attr_set_transform_write(self, stransform_write):
        if not stransform_write:
            self._transform_write = None
            return

        try:
            self.debug('parsing expression "%s"', stransform_write)
            transform_write = core_expressions.parse(self.get_id(), stransform_write)

            deps = transform_write.get_deps()
            for dep in deps:
                if not dep.startswith('$'):
                    continue

                if dep != '${}'.format(self._id):
                    raise core_expressions.ExpressionError('transform expression depends other ports')

            self.debug('setting write transform "%s"', transform_write)
            self._transform_write = transform_write

        except core_expressions.ExpressionError as e:
            self.error('failed to set transform write expression "%s": %s', stransform_write, e)

            raise InvalidAttributeValue('transform_write') from e

    async def read_value(self):
        raise NotImplementedError()

    async def write_value(self, value):
        pass

    def get_value(self):
        return self._value

    def get_change_reason(self):
        return self._change_reason

    def reset_change_reason(self):
        self._change_reason = CHANGE_REASON_NATIVE

    async def _write_value_queued(self, value, reason):
        done = asyncio.get_running_loop().create_future()

        await self._write_value_queue.put((value, reason, done))

        # wait for actual write_value operation to be done
        await done

    async def _write_value_loop(self):
        while True:
            try:
                value, reason, done = await self._write_value_queue.get()

                try:
                    result = await self._write_value(value, reason)
                    if not done.done():
                        done.set_result(result)

                except Exception:
                    done.set_exception(sys.exc_info()[1])

                await asyncio.sleep(await self.get_attr('write_value_pause'))

            except asyncio.CancelledError:
                self.debug('write value task cancelled')
                break

    async def _write_value(self, value, reason):
        self._last_write_value_time = time.time()
        self._change_reason = reason

        await self.write_value(value)
        await main.update()

    async def set_value(self, value, reason):
        if self._transform_write:
            # temporarily set the port value to the new value,
            # so that the write transform expression takes the new
            # value into consideration when evaluating the result
            prev_value = self._value
            self._value = value
            value = await self.adapt_value_type(self._transform_write.eval())
            self._value = prev_value

        try:
            await self._write_value_queued(value, reason)
            self.debug('wrote value %s (reason=%s)', json_utils.dumps(value), reason)

        except Exception:
            self.error('failed to write value %s (reason=%s)', json_utils.dumps(value), reason, exc_info=True)

            raise

    def set_value_fire_and_forget(self, value, reason):
        asyncio.create_task(self.set_value(value, reason))

    def push_value(self, value, reason):
        already_scheduled = self._asap_value is not None
        self._asap_value = value

        if not already_scheduled:
            async def set_value():
                asap_value = self._asap_value
                self._asap_value = None
                await self.set_value(asap_value, reason)

            asyncio.create_task(set_value())

        else:
            self.debug('will set value %s asap', json_utils.dumps(value))

    async def read_transformed_value(self):
        value = await self.read_value()
        if value is None:
            return None

        if self._transform_read:
            # temporarily set the new value to the port,
            # so that the read transform expression works as expected

            old_value = self._value
            self._value = value
            value = await self.adapt_value_type(self._transform_read.eval())
            self._value = old_value

        return value

    async def adapt_value_type(self, value):
        return self.adapt_value_type_sync(await self.get_type(), await self.get_attr('integer'), value)

    @staticmethod
    def adapt_value_type_sync(typ, integer, value):
        if value is None:
            return None

        if typ == TYPE_BOOLEAN:
            return bool(value)

        else:
            # round the value if port accepts only integers
            if integer:
                return int(value)

            return float(value)

    async def set_sequence(self, values, delays, repeat):
        if self._sequence:
            self.debug('canceling current sequence')
            self._sequence.cancel()
            self._sequence = None

        if values:
            callback = functools.partial(self.set_value_fire_and_forget, reason=CHANGE_REASON_SEQUENCE)
            self._sequence = core_sequences.Sequence(values, delays, repeat, callback, self._on_sequence_finish)

            self.debug('installing sequence')
            self._sequence.start()

    async def _on_sequence_finish(self):
        self.debug('sequence finished')

        self._sequence = None
        if await self.is_persisted():
            await self.save()

    def heart_beat(self):
        pass

    def heart_beat_second(self):
        pass

    @staticmethod
    def add_timeout(timeout, callback, *args, **kwargs):
        callback = functools.partial(callback, *args, **kwargs)

        return main.loop.call_later(timeout / 1000.0, callback)

    @staticmethod
    def cancel_timeout(handle):
        handle.cancel()

    async def to_json(self):
        attrs = await self.get_attrs()

        if self._enabled:
            attrs['value'] = self._value

        else:
            attrs['value'] = None

        attrdefs = copy.deepcopy(self.ADDITIONAL_ATTRDEFS)
        for attrdef in attrdefs.values():
            attrdef.pop('pattern', None)

        attrs['definitions'] = attrdefs

        return attrs

    async def load(self):
        self.debug('loading persisted data')

        data = persist.get(self.PERSIST_COLLECTION, self.get_id()) or {}
        await self.load_from_data(data)

        self.set_loaded()
        self.initialize()

    async def load_from_data(self, data):
        attrs_start = ['enabled']  # these will be loaded first, in this order
        attrs_end = ['expression']  # these will be loaded last, in this order

        attr_items = data.items()
        attr_items = [a for a in attr_items if (a[0] not in attrs_start) and (a[0] not in attrs_end)]

        attr_items_start = []
        for n in attrs_start:
            v = data.get(n)
            if v is not None:
                attr_items_start.append((n, v))

        # sort the rest of the attributes alphabetically
        attr_items.sort(key=lambda i: i[0])

        attr_items_end = []
        for n in attrs_end:
            v = data.get(n)
            if v is not None:
                attr_items_end.append((n, v))

        attr_items = attr_items_start + attr_items + attr_items_end

        for name, value in attr_items:
            if name in ('id', 'value'):
                continue  # value is also among the persisted fields

            try:
                self.debug('loading %s = %s', name, json_utils.dumps(value))
                await self.set_attr(name, value)

            except Exception as e:
                self.error('failed to set attribute %s = %s: %s', name, json_utils.dumps(value), e)

        # value
        if await self.is_persisted() and data.get('value') is not None:
            self._value = data['value']
            self.debug('loaded value = %s', json_utils.dumps(self._value))

            if await self.is_writable():
                # write the just-loaded value to the port
                value = self._value
                if self._transform_write:
                    value = await self.adapt_value_type(self._transform_write.eval())

                await self.write_value(value)

        elif self.is_enabled():
            try:
                value = await self.read_transformed_value()
                if value is not None:
                    self._value = value
                    self.debug('read value = %s', json_utils.dumps(self._value))

            except Exception as e:
                self.error('failed to read value: %s', e, exc_info=True)

    async def save(self):
        if not self.is_loaded():
            return

        d = await self.prepare_for_save()

        self.debug('persisting data')
        persist.replace(self.PERSIST_COLLECTION, self._id, d)

    async def prepare_for_save(self):
        # value
        d = {
            'id': self.get_id()
        }

        if await self.is_persisted():
            d['value'] = self._value

        else:
            d['value'] = None

        # attributes
        for name in self.get_modifiable_attrs():
            v = await self.get_attr(name)
            if v is None:
                continue

            d[name] = v

        return d

    async def cleanup(self):
        # cancel sequence
        if self._sequence:
            self.debug('canceling current sequence')
            self._sequence.cancel()
            self._sequence = None

        self._write_value_task.cancel()
        await self._write_value_task

    def is_loaded(self):
        return self._loaded

    def set_loaded(self):
        self._loaded = True

    def remove(self, persisted_data=True):
        # cancel sequence
        if self._sequence:
            self.debug('canceling current sequence')
            self._sequence.cancel()
            self._sequence = None

        _ports.pop(self._id, None)

        if persisted_data:
            self.debug('removing persisted data')
            persist.remove(self.PERSIST_COLLECTION, filt={'id': self._id})

        self.trigger_remove()

        return True

    def trigger_add(self):
        event = core_events.PortAdd(self)
        core_sessions.push(event)
        core_events.handle_event(event)

    def trigger_remove(self):
        event = core_events.PortRemove(self)
        core_sessions.push(event)
        core_events.handle_event(event)

    def trigger_update(self):
        event = core_events.PortUpdate(self)
        core_sessions.push(event)
        core_events.handle_event(event)

    def trigger_value_change(self):
        event = core_events.ValueChange(self)
        core_sessions.push(event)
        core_events.handle_event(event)

    async def get_schema(self):
        if self._schema is None:
            self._schema = {
                'type': 'object',
                'properties': {},
                'additionalProperties': False
            }

            for name, attrdef in self.ATTRDEFS.items():
                if not attrdef.get('modifiable'):
                    continue

                if await self.get_attr(name) is None:
                    continue

                attr_schema = dict(attrdef)

                if attr_schema['type'] == 'string':
                    if 'min' in attr_schema:
                        attr_schema['minLength'] = attr_schema.pop('min')

                    if 'max' in attr_schema:
                        attr_schema['maxLength'] = attr_schema.pop('max')

                elif attr_schema['type'] == 'number':
                    if attr_schema.get('integer'):
                        attr_schema['type'] = 'integer'

                    if 'min' in attr_schema:
                        attr_schema['minimum'] = attr_schema.pop('min')

                    if 'max' in attr_schema:
                        attr_schema['maximum'] = attr_schema.pop('max')

                if 'choices' in attrdef:
                    attr_schema['enum'] = [c['value'] for c in attr_schema.pop('choices')]

                attr_schema.pop('modifiable', None)

                self._schema['properties'][name] = attr_schema

        return self._schema

    async def get_value_schema(self):
        if self._value_schema is None:
            self._value_schema = {}

            c = await self.get_attr('choices')
            if c is not None:
                self._value_schema['enum'] = [i['value'] for i in c]

            else:
                m = await self.get_attr('min')
                if m is not None:
                    self._value_schema['minimum'] = m

                m = await self.get_attr('max')
                if m is not None:
                    self._value_schema['maximum'] = m

                if await self.get_attr('integer'):
                    self._value_schema['type'] = 'integer'

                elif await self.get_type() == TYPE_BOOLEAN:
                    self._value_schema['type'] = 'boolean'

                else:  # assuming number
                    self._value_schema['type'] = 'number'

        return self._value_schema


class Port(BasePort, metaclass=abc.ABCMeta):
    def __init__(self, port_id):
        super().__init__(port_id)

        self._tag = ''
        self._persisted = False


async def load(port_settings):
    port_driver_classes = {}
    ports = []

    for port_spec in port_settings:
        driver = port_spec.pop('driver', None)
        if not driver:
            logger.error('ignoring port with no driver')
            continue

        if isinstance(driver, str):
            if driver not in port_driver_classes:
                try:
                    logger.debug('loading port driver %s', driver)
                    port_driver_classes[driver] = utils.load_attr(driver)

                except Exception as e:
                    logger.error('failed to load port driver %s: %s', driver, e, exc_info=True)

                    continue

            port_class = port_driver_classes[driver]

        else:
            port_class = driver

        port_class_desc = '{}.{}'.format(port_class.__module__, port_class.__name__)

        try:
            value = port_spec.pop('value', None)  # initial value
            port = port_class(**port_spec)
            if value is not None:
                port._value = value

            _ports[port.get_id()] = port
            ports.append(port)

            logger.debug('initialized %s (driver %s)', port, port_class_desc)

        except Exception as e:
            logger.error('failed to initialize port from driver %s: %s', port_class_desc, e, exc_info=True)

    for old_id, new_id in settings.port_mappings.items():
        if new_id in _ports:
            logger.error('cannot map port %s: new id already exists', old_id, new_id)
            continue

        port = _ports.get(old_id)
        if not port:
            logger.error('cannot map port %s to %s: no such port', old_id, new_id)
            continue

        try:
            port.map_id(new_id)

        except Exception as e:
            port.error('cannot map to %s: %s', new_id, e)

        _ports.pop(old_id)
        _ports[port.get_id()] = port

    for port in ports:
        try:
            await port.load()

        except Exception as e:
            logger.error('failed to load %s: %s', port, e, exc_info=True)

        port.trigger_add()

    return ports


async def load_one(cls, args):
    ports = await load([dict(driver=cls, **args)])
    if not ports:
        return None

    return ports[0]


def get(port_id):
    return _ports.get(port_id)


def all_ports():
    return _ports.values()


async def cleanup():
    for port in _ports.values():
        await port.cleanup()


def reset():
    logger.debug('clearing ports persisted data')
    persist.remove(BasePort.PERSIST_COLLECTION)




import logging
import redis

from qtoggleserver.persist import BaseDriver
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)


class DuplicateRecordId(redis.RedisError):
    pass


class RedisDriver(BaseDriver):
    def __init__(self, host, port, db, **kwargs):
        logger.debug('connecting to %s:%s/%s', host, port, db)

        self._client = redis.StrictRedis(host=host, port=port, db=db, encoding='utf8', decode_responses=True)

    def query(self, collection, fields, filt, limit):
        db_records = []

        if 'id' in filt:  # look for specific record id
            filt = dict(filt)
            _id = filt.pop('id')
            db_record = self._client.hgetall(self._make_record_key(collection, _id))

            # apply filter criteria
            if db_record and self._filter_matches(db_record, filt):
                db_record['id'] = _id
                db_records.append(db_record)

        else:
            # look through all records from this collection, iterating through set
            for _id in self._client.sscan_iter(self._make_set_key(collection)):
                # retrieve the db record
                db_record = self._client.hgetall(self._make_record_key(collection, _id))
                db_record['id'] = _id

                # apply filter criteria
                if self._filter_matches(db_record, filt):
                    db_records.append(db_record)

        # apply limit
        if limit is not None:
            db_records = db_records[:limit]

        # transform from db record and return
        return (self._record_from_db(dbr) for dbr in db_records)

    def insert(self, collection, record):
        # make sure we have an id
        record = dict(record)
        _id = record.pop('id', None)
        if _id is None:
            _id = self._get_next_id(collection)

        key = self._make_record_key(collection, _id)
        set_key = self._make_set_key(collection)

        # check for duplicates
        if self._client.sismember(set_key, _id):
            raise DuplicateRecordId(_id)

        # adapt the record to db
        db_record = self._record_to_db(record)

        # actually insert the record
        self._client.hmset(key, db_record)

        # add the id to set
        self._client.sadd(set_key, _id)

    def update(self, collection, record_part, filt):
        # adapt the record part to db
        db_record_part = self._record_to_db(record_part)

        modified_count = 0

        if 'id' in filt:
            filt = dict(filt)
            _id = filt.pop('id')
            key = self._make_record_key(collection, _id)

            # retrieve the db record
            db_record = self._client.hgetall(key)
            if db_record and self._filter_matches(db_record, filt):
                self._client.hmset(key, db_record_part)

            modified_count = 1

        else:  # no id in filt
            # look through all records from this collection, iterating through set
            for _id in self._client.sscan_iter(self._make_set_key(collection)):
                key = self._make_record_key(collection, _id)

                # retrieve the db record
                db_record = self._client.hgetall(key)

                # apply filter criteria
                if not self._filter_matches(db_record, filt):
                    continue

                # actually update the record
                self._client.hmset(key, db_record_part)

                modified_count += 1

        return modified_count

    def replace(self, collection, _id, record, upsert):
        # adapt the record to db
        new_db_record = self._record_to_db(record)
        new_db_record.pop('id', None)  # never add the id together with other fields

        key = self._make_record_key(collection, _id)
        old_db_record = self._client.hgetall(key)

        if not old_db_record and not upsert:
            return False  # no record found, no replacing

        # remove any existing record
        self._client.delete(key)

        # insert the new record
        self._client.hmset(key, new_db_record)

        # make sure the id is present in set
        self._client.sadd(self._make_set_key(collection), _id)

        return True

    def remove(self, collection, filt):
        removed_count = 0

        if 'id' in filt:
            filt = dict(filt)
            _id = filt.pop('id')
            key = self._make_record_key(collection, _id)
            db_record = self._client.hgetall(key)

            # actually remove the record
            if db_record and self._filter_matches(db_record, filt):
                self._client.delete(key)
                removed_count = 1

            # remove the id from set
            self._client.srem(self._make_set_key(collection), _id)

        else:  # no id in filt
            ids_to_remove = set()

            # look through all records from this collection, iterating through set
            for _id in self._client.sscan_iter(self._make_set_key(collection)):
                key = self._make_record_key(collection, _id)

                # retrieve the db record
                db_record = self._client.hgetall(key)

                # apply filter criteria
                if not self._filter_matches(db_record, filt):
                    continue

                # actually remove the record
                self._client.delete(key)

                # remember ids to remove from set
                ids_to_remove.add(_id)

                removed_count += 1

            # remove the ids from set
            for _id in ids_to_remove:
                self._client.srem(self._make_set_key(collection), _id)

        return removed_count

    def close(self):
        pass

    def _filter_matches(self, db_record, filt):
        for key, value in filt.items():
            try:
                if db_record[key] != self._value_to_db(value):
                    return False

            except KeyError:
                return False

        return True

    def _get_next_id(self, collection):
        return int(self._client.incr(self._make_sequence_key(collection)))

    @classmethod
    def _record_from_db(cls, db_record):
        return {k: (cls._value_from_db(v) if k != 'id' else v) for k, v in db_record.items()}

    @classmethod
    def _record_to_db(cls, record):
        return {k: (cls._value_to_db(v) if k != 'id' else v) for k, v in record.items()}

    @staticmethod
    def _value_to_db(value):
        return json_utils.dumps(value)

    @staticmethod
    def _value_from_db(value):
        return json_utils.loads(value)

    @staticmethod
    def _make_record_key(collection, _id):
        if _id:
            return '{}:{}'.format(collection, _id)

        else:
            return collection

    @staticmethod
    def _make_set_key(collection):
        return '{}-id-set'.format(collection)

    @staticmethod
    def _make_sequence_key(collection):
        return '{}-id-sequence'.format(collection)




import logging
import pymongo
import re

from bson import ObjectId

from qtoggleserver.persist import BaseDriver


logger = logging.getLogger(__name__)

_OBJECT_ID_RE = re.compile('^[0-9a-f]{24}$')


class MongoDriver(BaseDriver):
    def __init__(self, host, port, db, **kwargs):
        logger.debug('connecting to %s:%s/%s', host, port, db)

        self._client = pymongo.MongoClient(host, port, serverSelectionTimeoutMS=200)
        self._db = self._client[db]

    def query(self, collection, fields, filt, limit):
        if fields:
            fields = dict((f, 1) for f in fields)
            if 'id' in fields:
                fields['_id'] = 1

            else:
                fields['_id'] = 0

        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db(filt.pop('id'))

        q = self._db[collection].find(filt, fields)
        if limit is not None:
            q = q.limit(limit)

        return self._query_gen_wrapper(q)

    def insert(self, collection, record):
        if 'id' in record:
            record = dict(record)
            record['_id'] = self._id_to_db(record.pop('id'))

        return self._db[collection].insert_one(record).inserted_id

    def update(self, collection, record_part, filt):
        if 'id' in record_part:
            record_part = dict(record_part)
            record_part['_id'] = self._id_to_db(record_part.pop('id'))

        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db(filt.pop('id'))

        return self._db[collection].update_many(filt, {'$set': record_part}, upsert=False).modified_count

    def replace(self, collection, _id, record, upsert):
        record = dict(record)
        record.pop('id', None)
        record['_id'] = self._id_to_db(_id)

        return bool(self._db[collection].replace_one({'_id': record['_id']}, record, upsert=upsert).modified_count)

    def remove(self, collection, filt=None):
        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db(filt.pop('id'))

        return self._db[collection].delete_many(filt).deleted_count

    def close(self):
        logger.debug('disconnecting mongo client')

        self._client.close()

    @classmethod
    def _query_gen_wrapper(cls, q):
        for r in q:
            if '_id' in r:
                r['id'] = cls._id_from_db(r.pop('_id'))

            yield r

    @staticmethod
    def _id_to_db(_id):
        if isinstance(_id, str) and _OBJECT_ID_RE.match(_id):
            return ObjectId(_id)

        else:
            return _id

    @staticmethod
    def _id_from_db(_id):
        if isinstance(_id, ObjectId):
            return str(_id)

        else:
            return _id




import copy
import datetime
import hashlib
import logging
import re
import socket
import sys

from qtoggleserver import system
from qtoggleserver import version
from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils.cmd import run_set_cmd


logger = logging.getLogger(__name__)

STANDARD_ATTRDEFS = {
    'name': {
        'type': 'string',
        'modifiable': True,
        'persisted': True,
        'min': 1,
        'max': 32,
        'pattern': r'^[_a-zA-Z][_a-zA-Z0-9-]{0,31}$',
        'internal': True
    },
    'display_name': {
        'type': 'string',
        'modifiable': True,
        'max': 64,
        'internal': True
    },
    'version': {
        'type': 'string',
        'internal': True
    },
    'api_version': {
        'type': 'string',
        'internal': True
    },
    'vendor': {
        'type': 'string',
        'internal': True
    },
    'admin_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'internal': True
    },
    'normal_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'internal': True
    },
    'viewonly_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'internal': True
    },
    'flags': {
        'type': ['string'],
        'internal': True
    },
    'virtual_ports': {
        'type': 'number',
        'enabled': lambda: bool(settings.core.virtual_ports),
        'internal': True
    },
    'uptime': {
        'type': 'number',
        'internal': True
    },
    'date': {
        'type': 'string',
        'pattern': r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.date.has_date_support(),
        'internal': True
    },
    'timezone': {
        'type': 'string',
        'modifiable': True,
        'persisted': False,
        'choices': [{'value': zone} for zone in system.date.get_timezones()],
        'enabled': lambda: system.date.has_timezone_support(),
        'internal': False  # Having internal False here enables export of attribute definition (needed for choices)
    },
    'network_ip': {
        'type': 'string',
        'pattern': r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/\d{1,2}:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d{1,3}\.\d{1,'
                   r'3}\.\d{1,3}\.\d{1,3})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_network_ip_support(),
        'internal': True
    },
    'network_wifi': {
        'type': 'string',
        # TODO this regex should ignore escaped colons \:
        'pattern': r'^(([^:]{0,32}:?)|([^:]{0,32}:[^:]{0,64}:?)|([^:]{0,32}:[^:]{0,64}:[0-9a-fA-F]{12}))$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_network_wifi_support(),
        'internal': True
    }
}

# TODO generalize additional attributes using getters and setters

'''
ADDITIONAL_ATTRDEFS = {
    'attr1': {
        'display_name': 'Some Attribute Display Name',
        'description': 'Some attribute description',
        'type': 'number',
        'modifiable': True,
        'unit': 'seconds',
        'min': 1,
        'max': 100,
        'integer': True,
        'step': 5,
        'reconnect': False,
        'choices': [
            {'value': 2, 'display_name': 'Two'},
            {'value': 4, 'display_name': 'Four'}
        ]
    },
    ...
}'''
ADDITIONAL_ATTRDEFS = {
    'ui_theme': {
        'display_name': 'Interface Theme',
        'description': 'Sets the user interface theme.',
        'type': 'string',
        'modifiable': True,
        'choices': [
            {'value': 'light', 'display_name': 'Light'},
            {'value': 'dark', 'display_name': 'Dark'}
        ]
    }
}

ATTRDEFS = dict(STANDARD_ATTRDEFS, **ADDITIONAL_ATTRDEFS)

EMPTY_PASSWORD_HASH = hashlib.sha256(b'').hexdigest()


name = socket.gethostname()
display_name = ''
admin_password_hash = None
normal_password_hash = None
viewonly_password_hash = None

ui_theme = 'dark'

_schema = None


def get_schema():
    global _schema

    if not _schema:
        _schema = {
            'type': 'object',
            'properties': {},
            'additionalProperties': False
        }

        # noinspection PyShadowingNames
        for name, attrdef in ATTRDEFS.items():
            if not attrdef.get('modifiable'):
                continue

            attr_schema = dict(attrdef)

            enabled = attr_schema.pop('enabled', True)
            if not enabled or callable(enabled) and not enabled():
                continue

            if attr_schema['type'] == 'string':
                if 'min' in attr_schema:
                    attr_schema['minLength'] = attr_schema.pop('min')

                if 'max' in attr_schema:
                    attr_schema['maxLength'] = attr_schema.pop('max')

            elif attr_schema['type'] == 'number':
                if attr_schema.get('integer'):
                    attr_schema['type'] = 'integer'

                if 'min' in attr_schema:
                    attr_schema['minimum'] = attr_schema.pop('min')

                if 'max' in attr_schema:
                    attr_schema['maximum'] = attr_schema.pop('max')

            if 'choices' in attrdef:
                attr_schema['enum'] = [c['value'] for c in attr_schema.pop('choices')]

            attr_schema.pop('persisted', None)
            attr_schema.pop('modifiable', None)
            attr_schema.pop('internal', None)

            _schema['properties'][name] = attr_schema

    return _schema


def get_attrs():
    attrs = {
        'name': name,
        'display_name': display_name,
        'version': version.VERSION,
        'api_version': core_api.API_VERSION,
        'vendor': version.VENDOR,
        'uptime': system.uptime(),
        'ui_theme': ui_theme
    }

    # never disclose these ones
    attrs['admin_password'] = ''
    attrs['normal_password'] = ''
    attrs['viewonly_password'] = ''

    flags = []
    if settings.system.fwupdate_driver:
        flags.append('firmware')

    if settings.core.listen_support:
        flags.append('listen')

    if settings.slaves.enabled:
        flags.append('master')

    if settings.reverse.enabled:
        flags.append('reverse')

    if settings.core.sequences_support:
        flags.append('sequences')

    if settings.core.ssl_support:
        flags.append('ssl')

    if settings.webhooks.enabled:
        flags.append('webhooks')

    attrs['flags'] = flags

    if settings.core.virtual_ports:
        attrs['virtual_ports'] = settings.core.virtual_ports

    if system.date.has_date_support():
        attrs['date'] = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')

    if system.date.has_timezone_support():
        attrs['timezone'] = system.date.get_timezone()

    if system.net.has_network_wifi_support():
        wifi_config = system.net.get_wifi_config()
        if wifi_config['bssid']:
            wifi_config['bssid'] = wifi_config['bssid'].replace(':', '')
            attrs['network_wifi'] = '{}:{}:{}'.format(wifi_config['ssid'], wifi_config['psk'], wifi_config['bssid'])

        elif wifi_config['psk']:
            wifi_config['psk'] = wifi_config['psk'].replace('\\', '\\\\')
            wifi_config['psk'] = wifi_config['psk'].replace(':', '\\:')
            attrs['network_wifi'] = '{}:{}'.format(wifi_config['ssid'], wifi_config['psk'])

        elif wifi_config['ssid']:
            wifi_config['ssid'] = wifi_config['ssid'].replace('\\', '\\\\')
            wifi_config['ssid'] = wifi_config['ssid'].replace(':', '\\:')
            attrs['network_wifi'] = wifi_config['ssid']

        else:
            attrs['network_wifi'] = ''

    if system.net.has_network_ip_support():
        ip_config = system.net.get_ip_config()
        if ip_config['ip'] and ip_config['mask'] and ip_config['gw'] and ip_config['dns']:
            attrs['network_ip'] = '{}/{}:{}:{}'.format(ip_config['ip'], ip_config['mask'],
                                                       ip_config['gw'], ip_config['dns'])

        else:
            attrs['network_ip'] = ''

    return attrs


def set_attrs(attrs):
    core_device_attrs = sys.modules[__name__]

    reboot_required = False

    # noinspection PyShadowingNames
    for name, value in attrs.items():
        # a few attributes may carry sensitive information
        # treat them separately and do not log their values
        if name.count('password'):
            logger.debug('setting device attribute %s', name)

        elif name == 'network_wifi':
            logger.debug('setting device attribute %s = [hidden]', name)

        else:
            logger.debug('setting device attribute %s = %s', name, json_utils.dumps(value))

        attrdef = ATTRDEFS[name]

        if not attrdef.get('modifiable'):
            return 'attribute not modifiable: {}'.format(name)

        # Treat passwords separately, as they are not persisted as given, but hashed first
        if name.endswith('_password') and hasattr(core_device_attrs, name + '_hash'):
            # Call password set command, if available
            if settings.password_set_cmd:
                run_set_cmd(settings.password_set_cmd, cmd_name='password', log_values=False,
                            username=name[:-9], password=value)

            value = hashlib.sha256(value.encode()).hexdigest()
            name += '_hash'

            setattr(core_device_attrs, name, value)
            continue

        persisted = attrdef.get('persisted', attrdef.get('modifiable'))
        if callable(persisted):
            persisted = persisted()
        if persisted:
            setattr(core_device_attrs, name, value)

        if name == 'name' and settings.device_name.set_cmd:
            run_set_cmd(settings.device_name.set_cmd, cmd_name='device name', name=value)

        elif name == 'date' and system.date.has_date_support():
            try:
                date = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')

            except ValueError:
                return 'invalid field: {}'.format(name)

            system.date.set_date(date)

        elif name == 'timezone' and system.date.has_timezone_support():
            system.date.set_timezone(value)

        elif name == 'network_wifi' and system.net.has_network_wifi_support():
            parts = value.split(':')
            i = 0
            while i < len(parts):
                if len(parts[i]) and parts[i][-1] == '\\':
                    parts[i] = parts[i][:-1] + ':' + parts[i + 1]
                    del parts[i + 1]
                i += 1

            parts = [p.replace('\\\\', '\\') for p in parts]
            while len(parts) < 3:
                parts.append('')

            ssid, psk, bssid = parts[:3]
            bssid = bssid.lower()
            bssid = re.sub('([a-f0-9]{2})', '\\1:', bssid).strip(':')  # Add colons

            system.net.set_wifi_config(ssid, psk, bssid)
            reboot_required = True

        elif name == 'network_ip' and system.net.has_network_ip_support():
            if value:
                parts = value.split(':')
                ip_mask, gw, dns = parts
                ip, mask = ip_mask.split('/')
                system.net.set_ip_config(ip, mask, gw, dns)

            else:
                system.net.set_ip_config(ip='', mask='', gw='', dns='')

            reboot_required = True

    return reboot_required


def to_json():
    attrdefs = copy.deepcopy(ATTRDEFS)
    filtered_attrdefs = {}
    for n, attrdef in attrdefs.items():
        if attrdef.pop('internal', False):
            continue

        enabled = attrdef.pop('enabled', True)
        if not enabled or callable(enabled) and not enabled():
            continue

        attrdef.pop('persisted', None)
        attrdef.pop('pattern', None)

        filtered_attrdefs[n] = attrdef

    result = get_attrs()
    result['definitions'] = filtered_attrdefs

    return result




import abc
import asyncio
import functools
import queue
import logging
import threading

from qtoggleserver import utils
from qtoggleserver.core import ports as core_ports

from . import add_cleanup_hook


logger = logging.getLogger(__name__)


class RunnerBusy(Exception):
    pass


class ThreadedRunner(threading.Thread, metaclass=abc.ABCMeta):
    QUEUE_TIMEOUT = 1

    def __init__(self, queue_size=None):
        self._running = False
        self._loop = asyncio.get_event_loop()
        self._queue = queue.Queue(queue_size or 0)
        self._queue_size = queue_size
        self._stopped_future = self._loop.create_future()

        super().__init__()

    def run(self):
        while self._running:
            try:
                func, callback = self._queue.get(timeout=self.QUEUE_TIMEOUT)

            except queue.Empty:
                continue

            try:
                result = func()

            except Exception as e:
                self._loop.call_soon_threadsafe(callback, None, e)

            else:
                self._loop.call_soon_threadsafe(callback, result, None)

        self._loop.call_soon_threadsafe(self._stopped_future.set_result, None)

    def schedule_func(self, func, callback):
        try:
            self._queue.put_nowait((func, callback))

        except queue.Full:
            raise RunnerBusy() from None

    def is_running(self):
        return self._running

    def start(self):
        self._running = True
        super().start()

    def stop(self):
        self._running = False

        return self._stopped_future


class Peripheral(utils.ConfigurableMixin, utils.LoggableMixin, metaclass=abc.ABCMeta):
    RUNNER_CLASS = ThreadedRunner
    RUNNER_QUEUE_SIZE = 8

    logger = logger
    _peripherals_by_address = {}

    @classmethod
    def get(cls, address, name, **kwargs):
        if name is None:
            name = ''

        if address not in cls._peripherals_by_address:
            logger.debug('initializing peripheral %s(%s@%s)', cls.__name__, name, address)
            peripheral = cls.make_peripheral(address, name, **kwargs)
            cls._peripherals_by_address[address] = peripheral

        return cls._peripherals_by_address[address]

    @classmethod
    def make_peripheral(cls, address, name, **kwargs):
        return cls(address, name, **kwargs)

    def __init__(self, address, name, **kwargs):
        utils.LoggableMixin.__init__(self, name, self.logger)

        self._address = address
        self._name = name
        self._ports = []
        self._enabled = False
        self._runner = None

        add_cleanup_hook(self.handle_cleanup)

    def __str__(self):
        return self._name

    def get_address(self):
        return self._address

    def get_name(self):
        return self._name

    def add_port(self, port):
        self._ports.append(port)

    def get_ports(self):
        return list(self._ports)

    def trigger_port_update(self):
        for port in self._ports:
            port.trigger_update()

    def is_enabled(self):
        return self._enabled

    async def enable(self):
        if self._enabled:
            return

        self._enabled = True
        await self.handle_enable()
        self.debug('peripheral enabled')

    async def disable(self):
        if not self._enabled:
            return

        self._enabled = False
        await self.handle_disable()
        self.debug('peripheral disabled')

    async def check_disabled(self, exclude_port=None):
        if not self._enabled:
            return

        for port in self._ports:
            if port.is_enabled() and port != exclude_port:
                break

        else:
            self.debug('all ports are disabled, disabling peripheral')
            await self.disable()

    def get_runner(self):
        if self._runner is None:
            self._runner = self.make_runner()

        return self._runner

    def make_runner(self):
        self.debug('starting threaded runner')
        runner = self.RUNNER_CLASS(queue_size=self.RUNNER_QUEUE_SIZE)
        runner.start()

        return runner

    async def run_threaded(self, func, *args, **kwargs):
        future = asyncio.get_running_loop().create_future()

        def callback(result, exception):
            if exception:
                future.set_exception(exception)

            else:
                future.set_result(result)

        runner = self.get_runner()
        runner.schedule_func(functools.partial(func, *args, **kwargs), callback)

        return await future

    async def handle_enable(self):
        pass

    async def handle_disable(self):
        pass

    async def handle_cleanup(self):
        if self._runner:
            self.debug('stopping threaded runner')
            await self._runner.stop()
            self.debug('threaded runner stopped')


class PeripheralPort(core_ports.Port, metaclass=abc.ABCMeta):
    PERIPHERAL_CLASS = Peripheral
    ID = 'port'

    ADDITIONAL_ATTRDEFS = {
        'address': {
            'display_name': 'Address',
            'description': 'The peripheral address.',
            'type': 'string',
            'modifiable': False
        }
    }

    def __init__(self, address, peripheral_name=None, **kwargs):
        self._peripheral = self.PERIPHERAL_CLASS.get(address, peripheral_name, **kwargs)
        self._peripheral.add_port(self)

        _id = self.make_id()
        if self._peripheral.get_name():
            _id = '{}.{}'.format(self._peripheral.get_name(), _id)

        super().__init__(_id)

    def make_id(self):
        return self.ID

    def get_peripheral(self):
        return self._peripheral

    async def attr_get_address(self):
        return self._peripheral.get_address()

    async def handle_enable(self):
        if not self._peripheral.is_enabled():
            await self._peripheral.enable()

    async def handle_disable(self):
        if self._peripheral.is_enabled():
            await self._peripheral.check_disabled(self)




import datetime
import json
import jsonpointer


def _resolve_refs_rec(obj, root_obj):
    if isinstance(obj, dict):
        if len(obj.keys()) == 1 and list(obj.keys())[0] == '$ref':
            ref = list(obj.values())[0]
            ref = ref[1:]  # Skip starting hash
            return jsonpointer.resolve_pointer(root_obj, ref)

        for k, v in obj.items():
            obj[k] = _resolve_refs_rec(v, root_obj)

    elif isinstance(obj, list):
        for i, e in enumerate(obj):
            obj[i] = _resolve_refs_rec(e, root_obj)

    return obj


def _make_json_encoder(date_format='%Y-%m-%d', datetime_format='%Y-%m-%dT%H:%M:%SZ'):
    def encode_default_json(obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime(datetime_format)

        elif isinstance(obj, datetime.date):
            return obj.strftime(date_format)

        elif isinstance(obj, (set, tuple)):
            return list(obj)

    return encode_default_json


def dumps(obj):
    if isinstance(obj, str):
        return '"' + obj + '"'

    elif isinstance(obj, bool):
        return ['false', 'true'][obj]

    elif isinstance(obj, (int, float)):
        return str(obj)

    elif obj is None:
        return 'null'

    else:
        return json.dumps(obj, default=_make_json_encoder())


def loads(s, resolve_refs=False):
    obj = json.loads(s)

    if resolve_refs:
        obj = _resolve_refs_rec(obj, root_obj=obj)

    return obj




import abc
import logging

from qtoggleserver.core import expressions as core_expressions

from ..types import device as device_events
from ..types import port as port_events
from ..types import slave as slave_events


logger = logging.getLogger(__package__)


class BaseEventHandler(metaclass=abc.ABCMeta):
    logger = logger

    # noinspection PyShadowingBuiltins
    def __init__(self, filter=None):
        self._filter = filter or {}
        self._filter_prepared = False
        self._filter_event_types = None
        self._filter_device_attrs = {}
        self._filter_port_value = None
        self._filter_port_attrs = {}

        # Maintain an internal "last" state for all objects, so we can detect changes in various attributes
        self._device_attrs = {}
        self._port_values = {}
        self._port_attrs = {}
        self._slave_attrs = {}

    def _prepare_filter(self):
        event_types = self._filter.get('type')
        if event_types:
            if not isinstance(event_types, list):
                event_types = event_types,
            self._filter_event_types = set(event_types)

        port_value = self._filter.get('port_value')
        if port_value is not None:
            if isinstance(port_value, str):  # An expression
                try:
                    self.logger.debug('using value expression "%s"', port_value)
                    self._filter_port_value = core_expressions.parse(self_port_id=None, sexpression=port_value)

                except core_expressions.ExpressionError as e:
                    self.logger.error('failed to parse port expression "%s": %s', port_value, e)

                    raise

            else:
                self._filter_port_value = port_value

        self._filter_device_attrs = {n[7:]: v for n, v in self._filter.items()
                                     if n.startswith('device_')}

        self._filter_port_attrs = {n[5:]: v for n, v in self._filter.items()
                                   if n.startswith('port_') and n != 'port_value'}

        self._filter_slave_attrs = {n[6:]: v for n, v in self._filter.items()
                                    if n.startswith('slave_')}

        self._filter_prepared = True

    @staticmethod
    def _make_changed_added_removed(old_attrs, new_attrs):
        changed_attrs = {}
        added_attrs = {}
        removed_attrs = {}

        all_attr_names = set(old_attrs) | set(new_attrs)
        for n in all_attr_names:
            old_v = old_attrs.get(n)
            new_v = new_attrs.get(n)
            if old_v == new_v:
                continue

            if old_v is None:
                if new_v is not None:
                    added_attrs[n] = new_v

            elif new_v is None:
                removed_attrs[n] = old_v

            else:
                changed_attrs[n] = (old_v, new_v)

        return changed_attrs, added_attrs, removed_attrs

    async def _update_from_event(self, event):
        value_pair = (None, None)
        old_attrs = {}
        new_attrs = {}
        changed_attrs = {}
        added_attrs = {}
        removed_attrs = {}

        if isinstance(event, port_events.PortEvent):
            port = event.get_port()

            old_attrs = self._port_attrs.get(port.get_id(), {})
            new_attrs = await port.get_attrs()

            old_value = self._port_values.get(port.get_id())
            new_value = port.get_value()
            value_pair = (old_value, new_value)

            if isinstance(event, (port_events.PortAdd, port_events.PortUpdate)):
                changed_attrs, added_attrs, removed_attrs = self._make_changed_added_removed(old_attrs, new_attrs)
                self._port_values[port.get_id()] = new_value
                self._port_attrs[port.get_id()] = new_attrs

            elif isinstance(event, port_events.PortRemove):
                self._port_values.pop(port.get_id(), None)
                removed_attrs = self._port_attrs.pop(port.get_id(), {})

            elif isinstance(event, port_events.ValueChange):
                self._port_values[port.get_id()] = new_value

        elif isinstance(event, device_events.DeviceEvent):
            old_attrs = self._device_attrs
            new_attrs = event.get_attrs()

            changed_attrs, added_attrs, removed_attrs = self._make_changed_added_removed(old_attrs, new_attrs)
            self._device_attrs = new_attrs

        elif isinstance(event, slave_events.SlaveDeviceEvent):
            slave = event.get_slave()
            slave_json = slave.to_json()

            # Flatten slave master properties and attributes
            old_attrs = self._slave_attrs.get(slave.get_name(), {})
            new_attrs = dict(slave_json, **slave_json.pop('attrs'))

            if isinstance(event, (slave_events.SlaveDeviceAdd, slave_events.SlaveDeviceUpdate)):
                changed_attrs, added_attrs, removed_attrs = self._make_changed_added_removed(old_attrs, new_attrs)
                self._slave_attrs[slave.get_name()] = new_attrs

            elif isinstance(event, slave_events.SlaveDeviceRemove):
                removed_attrs = self._slave_attrs.pop(slave.get_name(), {})

        return value_pair, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs

    @staticmethod
    def _accepts_attrs(filter_attrs, old_attrs, new_attrs):
        for name, filter_value in filter_attrs.items():
            if isinstance(filter_value, list):
                old_filter_value, filter_value = filter_value[:2]
                if old_filter_value != old_attrs.get(name):
                    return False

            if filter_value != new_attrs.get(name):
                return False

        return True

    async def accepts_device(self, event, old_attrs, new_attrs):
        return self._accepts_attrs(self._filter_device_attrs, old_attrs, new_attrs)

    async def accepts_port_value(self, event, value_pair):
        old_value, new_value = value_pair
        filter_value = self._filter_port_value

        if isinstance(filter_value, list):
            old_filter_value, filter_value = filter_value[:2]
            if old_filter_value != old_value:
                return False

        if isinstance(filter_value, core_expressions.Expression):
            filter_value = filter_value.eval()

        if new_value != filter_value:
            return False

        return True

    async def accepts_port(self, event, value_pair, old_attrs, new_attrs):
        if self._filter_port_value is not None:
            if not await self.accepts_port_value(event, value_pair):
                return False

        return self._accepts_attrs(self._filter_port_attrs, old_attrs, new_attrs)

    async def accepts_slave(self, event, old_attrs, new_attrs):
        return self._accepts_attrs(self._filter_slave_attrs, old_attrs, new_attrs)

    async def accepts(self, event, value_pair, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs):
        if not self._filter_prepared:
            self._prepare_filter()

        if self._filter_event_types and event.get_type() not in self._filter_event_types:
            return False

        if (isinstance(event, device_events.DeviceEvent) and
            not await self.accepts_device(event, old_attrs, new_attrs)):

            return False

        elif (isinstance(event, port_events.PortEvent) and
              not await self.accepts_port(event, value_pair, old_attrs, new_attrs)):

            return False

        elif (isinstance(event, slave_events.SlaveDeviceEvent) and
              not await self.accepts_slave(event, old_attrs, new_attrs)):

            return False

        return True

    async def handle_event(self, event):
        (value_pair, old_attrs, new_attrs,
         changed_attrs, added_attrs, removed_attrs) = await self._update_from_event(event)

        accepted = await self.accepts(event, value_pair, old_attrs, new_attrs,
                                      changed_attrs, added_attrs, removed_attrs)

        if not accepted:
            self.logger.debug('skipping event %s', event)
            return

        self.logger.debug('handling event %s', event)

        try:
            await self.on_event(event)

        except Exception as e:
            self.logger.error('failed to handle event %s: %s', event, e, exc_info=True)

        try:
            if isinstance(event, port_events.ValueChange):
                old_value, new_value = value_pair
                await self.on_value_change(event, event.get_port(), old_value, new_value, new_attrs)

            elif isinstance(event, port_events.PortUpdate):
                await self.on_port_update(event, event.get_port(), old_attrs, new_attrs,
                                          changed_attrs, added_attrs, removed_attrs)

            elif isinstance(event, port_events.PortAdd):
                await self.on_port_add(event, event.get_port(), new_attrs)

            elif isinstance(event, port_events.PortRemove):
                await self.on_port_remove(event, event.get_port(), new_attrs)

            elif isinstance(event, device_events.DeviceUpdate):
                await self.on_device_update(event, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs)

            elif isinstance(event, slave_events.SlaveDeviceUpdate):
                await self.on_slave_device_update(event, event.get_slave(), old_attrs, new_attrs,
                                                  changed_attrs, added_attrs, removed_attrs)

            elif isinstance(event, slave_events.SlaveDeviceAdd):
                await self.on_slave_device_add(event, event.get_slave(), new_attrs)

            elif isinstance(event, slave_events.SlaveDeviceRemove):
                await self.on_slave_device_remove(event, event.get_slave(), new_attrs)

        except Exception as e:
            self.logger.error('failed to handle event %s: %s', event, e, exc_info=True)

    async def on_event(self, event):
        pass

    async def on_value_change(self, event, port, old_value, new_value, attrs):
        pass

    async def on_port_update(self, event, port, old_attrs, new_attrs,
                             changed_attrs, added_attrs, removed_attrs):
        pass

    async def on_port_add(self, event, port, attrs):
        pass

    async def on_port_remove(self, event, port, attrs):
        pass

    async def on_device_update(self, event, old_attrs, new_attrs,
                               changed_attrs, added_attrs, removed_attrs):
        pass

    async def on_slave_device_update(self, event, slave, old_attrs, new_attrs,
                                     changed_attrs, added_attrs, removed_attrs):
        pass

    async def on_slave_device_add(self, event, slave, attrs):
        pass

    async def on_slave_device_remove(self, event, slave, attrs):
        pass




import abc
import datetime
import math
import time


FUNCTIONS = {}


def function(name):
    def decorator(func_class):
        func_class.NAME = name
        FUNCTIONS[name] = func_class

        return func_class

    return decorator


class ExpressionError(Exception):
    pass


class InvalidExpression(ExpressionError):
    pass


class InvalidArgument(ExpressionError):
    def __init__(self, arg_no, value):
        self.arg_no = arg_no
        self.value = value

    def __str__(self):
        return 'invalid argument {}: {}'.format(self.arg_no, self.value)


class CircularDependency(ExpressionError):
    pass


class IncompleteExpression(ExpressionError):
    pass


class Expression(metaclass=abc.ABCMeta):
    def eval(self):
        raise NotImplementedError()

    def get_deps(self):
        # special deps:
        #  * 'time' - used to indicate dependency on system time (seconds)
        #  * 'time_ms' - used to indicate dependency on system time (milliseconds)

        return set()

    @staticmethod
    def parse(self_port_id, sexpression):
        raise NotImplementedError


class Constant(Expression):
    def __init__(self, value, sexpression):
        self.value = value
        self.sexpression = sexpression

    def __str__(self):
        return self.sexpression

    def eval(self):
        return self.value

    @staticmethod
    def parse(self_port_id, sexpression):
        sexpression = sexpression.strip()

        if sexpression == 'true':
            value = 1

        elif sexpression == 'false':
            value = 0

        else:
            try:
                value = int(sexpression)

            except ValueError:
                try:
                    value = float(sexpression)

                except ValueError:
                    raise InvalidExpression('"{}" is not a valid constant'.format(sexpression)) from None

        return Constant(value, sexpression)


class Function(Expression, metaclass=abc.ABCMeta):
    NAME = None
    MIN_ARGS = None
    MAX_ARGS = None

    def __init__(self, args):
        self.args = args
        self._deps = None

    def __str__(self):
        s = getattr(self, '_str', None)
        if s is None:
            args_str = ', '.join([str(e) for e in self.args])
            self._str = s = '{}({})'.format(self.NAME, args_str)

        return s

    def get_deps(self):
        if self._deps is None:
            self._deps = set()

            for arg in self.args:
                self._deps |= arg.get_deps()

        return self._deps

    def eval_args(self):
        return [a.eval() for a in self.args]

    @staticmethod
    def parse(self_port_id, sexpression):
        sexpression = sexpression.strip()

        p_start = None
        p_end = None
        p_last_comma = None
        level = 0
        sargs = []
        for i, c in enumerate(sexpression):
            if c == '(':
                if p_start is None:
                    p_start = i

                elif level == 0:
                    raise InvalidExpression('empty function call')

                level += 1

            elif c == ')':
                if level == 0:
                    raise InvalidExpression('unbalanced parentheses')

                elif level == 1:
                    if (p_end is None) and (i == len(sexpression) - 1):
                        p_end = i

                    else:
                        raise InvalidExpression('unexpected text after function call')

                level -= 1

            elif c == ',' and level == 1:
                sargs.append(sexpression[(p_last_comma or p_start) + 1: i])
                p_last_comma = i

        if (p_start is None) or (p_end is None) or (p_start > p_end) or (level != 0):
            raise InvalidExpression('unbalanced parentheses')

        if p_end - p_start > 1:
            sargs.append(sexpression[(p_last_comma or p_start) + 1: p_end])

        func_name = sexpression[:p_start].strip()
        func_class = FUNCTIONS.get(func_name)
        if func_class is None:
            raise InvalidExpression('unknown function "{}"'.format(func_name))

        if func_class.MIN_ARGS is not None and len(sargs) < func_class.MIN_ARGS:
            raise InvalidExpression('too few arguments for function "{}"'.format(func_name))

        if func_class.MAX_ARGS is not None and len(sargs) > func_class.MAX_ARGS:
            raise InvalidExpression('too many arguments for function "{}"'.format(func_name))

        args = [parse(self_port_id, sa) for sa in sargs]

        return func_class(args)


@function('ADD')
class AddFunction(Function):
    MIN_ARGS = 2

    def eval(self):
        return sum(self.eval_args())


@function('SUB')
class SubFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()
        return eval_args[0] - eval_args[1]


@function('MUL')
class MulFunction(Function):
    MIN_ARGS = 2

    def eval(self):
        r = 1
        for e in self.eval_args():
            r *= e

        return r


@function('DIV')
class DivFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        if eval_args[1]:
            return int(eval_args[0] / eval_args[1])

        else:
            return 0


@function('MOD')
class ModFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        if eval_args[1]:
            return eval_args[0] % eval_args[1]

        else:
            return 0


@function('AND')
class AndFunction(Function):
    MIN_ARGS = 2

    def eval(self):
        r = True
        for e in self.eval_args():
            r = r and bool(e)

        return int(r)


@function('OR')
class OrFunction(Function):
    MIN_ARGS = 2

    def eval(self):
        r = False
        for e in self.eval_args():
            r = r or bool(e)

        return int(r)


@function('NOT')
class NotFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self):
        return int(not bool(self.eval_args()[0]))


@function('XOR')
class XorFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        e1 = bool(eval_args[0])
        e2 = bool(eval_args[1])

        return int(e1 and not e2 or e2 and not e1)


@function('BITAND')
class BitAndFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        r = -1
        for e in self.eval_args():
            r &= int(e)

        return r


@function('BITOR')
class BitOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        r = 0
        for e in self.eval_args():
            r |= int(e)

        return r


@function('BITNOT')
class BitNotFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self):
        return ~int(self.eval_args()[0])


@function('BITXOR')
class BitXOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0]) ^ int(eval_args[1])


@function('SHL')
class SHLFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0]) << int(eval_args[1])


@function('SHR')
class SHRFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0]) >> int(eval_args[1])


@function('IF')
class IfFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def eval(self):
        eval_args = self.eval_args()

        if eval_args[0]:
            return eval_args[1]

        else:
            return eval_args[2]


@function('EQ')
class EqFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0] == eval_args[1])


@function('GT')
class GTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0] > eval_args[1])


@function('GTE')
class GTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0] >= eval_args[1])


@function('LT')
class LTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0] < eval_args[1])


@function('LTE')
class LTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        return int(eval_args[0] <= eval_args[1])


@function('ABS')
class AbsFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self):
        return abs(self.eval_args()[0])


@function('SGN')
class SgnFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self):
        e = int(self.eval_args()[0])
        if e > 0:
            return 1

        elif e < 0:
            return -1

        else:
            return 0


@function('MIN')
class MinFunction(Function):
    MIN_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        m = eval_args[0]
        for e in eval_args[1:]:
            if e < m:
                m = e

        return m


@function('MAX')
class MaxFunction(Function):
    MIN_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        m = eval_args[0]
        for e in eval_args[1:]:
            if e > m:
                m = e

        return m


@function('FLOOR')
class FloorFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self):
        eval_args = self.eval_args()

        return int(math.floor(eval_args[0]))


@function('CEIL')
class CeilFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self):
        eval_args = self.eval_args()

        return int(math.ceil(eval_args[0]))


@function('ROUND')
class RoundFunction(Function):
    MIN_ARGS = 1
    MAX_ARGS = 2

    def eval(self):
        eval_args = self.eval_args()

        v = eval_args[0]
        d = 0
        if len(eval_args) == 2:
            d = eval_args[1]

        return round(v, d)


@function('TIME')
class TimeFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return int(time.time())


@function('TIMEMS')
class TimeMSFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time_ms'}

    def eval(self):
        return int(time.time() * 1000)


@function('HELD')
class HeldFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._time_ms = None
        self._last_value = None

    def get_deps(self):
        return {'time_ms'}

    def eval(self):
        time_ms = int(time.time() * 1000)
        result = False

        value = self.args[0].eval()
        fixed_value = self.args[1].eval()
        duration = self.args[2].eval()

        if self._time_ms is None:  # very first expression eval call
            self._time_ms = time_ms

        else:
            delta = time_ms - self._time_ms

            if self._last_value != value:
                self._time_ms = time_ms  # reset held timer

            else:
                result = (delta >= duration) and (value == fixed_value)

        self._last_value = value

        return result


@function('DELAY')
class DelayFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    HISTORY_SIZE = 1024

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._queue = []
        self._last_value = None
        self._current_value = None

    def get_deps(self):
        return {'time_ms'}

    def eval(self):
        time_ms = int(time.time() * 1000)

        value = self.args[0].eval()
        delay = self.args[1].eval()

        if self._current_value is None:
            self._current_value = value

        # detect value transitions and build history
        if value != self._last_value:
            self._last_value = value

            # drop elements from queue if history size reached
            while len(self._queue) >= self.HISTORY_SIZE:
                self._queue.pop(0)

            self._queue.append((time_ms, value))

        # process history
        while self._queue and (time_ms - self._queue[0][0]) >= delay:
            self._current_value = self._queue.pop(0)[1]

        return self._current_value


@function('HYST')
class HystFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._last_result = 0

    def eval(self):
        value = self.args[0].eval()
        threshold1 = self.args[1].eval()
        threshold2 = self.args[2].eval()

        self._last_result = int((self._last_result == 0 and value > threshold1) or
                                (self._last_result != 0 and value >= threshold2))

        return self._last_result


@function('YEAR')
class YearFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return datetime.datetime.now().year


@function('MONTH')
class MonthFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return datetime.datetime.now().month


@function('DAY')
class DayFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return datetime.datetime.now().day


@function('DOW')
class DOWFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return datetime.datetime.now().weekday()


@function('HOUR')
class HourFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return datetime.datetime.now().hour


@function('MINUTE')
class MinuteFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return datetime.datetime.now().minute


@function('SECOND')
class SecondFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self):
        return {'time'}

    def eval(self):
        return datetime.datetime.now().second


@function('HMSINTERVAL')
class HMSIntervalFunction(Function):
    MIN_ARGS = 4
    MAX_ARGS = 6

    def get_deps(self):
        return {'time'}

    def eval(self):
        now = datetime.datetime.now()

        if len(self.args) >= 6:
            start_h = self.args[0].eval()
            start_m = self.args[1].eval()
            start_s = self.args[2].eval()
            stop_h = self.args[3].eval()
            stop_m = self.args[4].eval()
            stop_s = self.args[5].eval()

            if not (0 <= start_h <= 23):
                raise InvalidArgument(1, start_h)

            if not (0 <= start_m <= 59):
                raise InvalidArgument(2, start_m)

            if not (0 <= start_s <= 59):
                raise InvalidArgument(3, start_s)

            if not (0 <= stop_h <= 23):
                raise InvalidArgument(4, stop_h)

            if not (0 <= stop_m <= 59):
                raise InvalidArgument(5, stop_m)

            if not (0 <= stop_s <= 59):
                raise InvalidArgument(6, stop_s)

            start_time = datetime.time(start_h, start_m, start_s)
            stop_time = datetime.time(stop_h, stop_m, stop_s)

        else:  # assuming 4
            start_h = self.args[0].eval()
            start_m = self.args[1].eval()
            stop_h = self.args[2].eval()
            stop_m = self.args[3].eval()

            if not (0 <= start_h <= 23):
                raise InvalidArgument(1, start_h)

            if not (0 <= start_m <= 59):
                raise InvalidArgument(2, start_m)

            if not (0 <= stop_h <= 23):
                raise InvalidArgument(3, stop_h)

            if not (0 <= stop_m <= 59):
                raise InvalidArgument(4, stop_m)

            start_time = datetime.time(start_h, start_m, 00)
            stop_time = datetime.time(stop_h, stop_m, 59)

        start_dt = datetime.datetime.combine(now.date(), start_time)
        stop_dt = datetime.datetime.combine(now.date(), stop_time)

        return start_dt <= now <= stop_dt


class PortValue(Expression):
    def __init__(self, port_id):
        self.port_id = port_id

    def __str__(self):
        return '${}'.format(self.port_id)

    def get_deps(self):
        return {'${}'.format(self.port_id)}

    def get_port(self):
        from qtoggleserver.core import ports as core_ports

        return core_ports.get(self.port_id)

    def eval(self):
        port = self.get_port()
        if not port:
            raise IncompleteExpression('unknown port {}'.format(self.port_id))

        if not port.is_enabled():
            raise IncompleteExpression('{} is disabled'.format(port))

        value = port.get_value()
        if value is None:
            raise IncompleteExpression('value of port {} is undefined'.format(port))

        return float(value)

    @staticmethod
    def parse(self_port_id, sexpression):
        sexpression = sexpression.strip()
        port_id = sexpression.strip('$')

        if port_id:
            return PortValue(port_id)

        else:
            return SelfPortValue(self_port_id)


class SelfPortValue(PortValue):
    def __str__(self):
        return '$'


def parse(self_port_id, sexpression):
    sexpression = sexpression.strip()
    if sexpression.startswith('$'):
        return PortValue.parse(self_port_id, sexpression)

    elif sexpression.count('('):
        return Function.parse(self_port_id, sexpression)

    else:
        return Constant.parse(self_port_id, sexpression)


async def check_loops(port, expression):
    seen_ports = {port}

    async def check_loops_rec(level, e):
        if isinstance(e, PortValue):
            p = e.get_port()
            if not p:
                return 0

            # a loop is detected when we stumble upon the initial port at a level deeper than 1
            if port is p and level > 1:
                return level

            # avoid visiting the same port twice
            if p in seen_ports:
                return 0

            seen_ports.add(p)

            expr = await p.get_expression()
            if expr:
                lv = await check_loops_rec(level + 1, expr)
                if lv:
                    return lv

            return 0

        elif isinstance(e, Function):
            for arg in e.args:
                lv = await check_loops_rec(level, arg)
                if lv:
                    return lv

        return 0

    if await check_loops_rec(1, expression) > 1:
        raise CircularDependency('{} is recursively referred'.format(port))




import inspect
import logging

from tornado.web import RequestHandler, HTTPError, StaticFileHandler as TornadoStaticFileHandler
from tornado.iostream import StreamClosedError

from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.core.api import funcs as core_api_funcs
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.slaves.api import funcs as slaves_api_funcs
from qtoggleserver.ui.api import funcs as ui_api_funcs
from qtoggleserver.utils import json as json_utils

from .constants import FRONTEND_URL_PREFIX
from .j2template import J2TemplateMixin
from .quicontext import make_context


logger = logging.getLogger(__name__)


class NoSuchFunction(HTTPError):
    def __init__(self):
        super().__init__(404, 'no such function')


class BaseHandler(RequestHandler):
    _UNDEFINED = {}

    def __init__(self, *args, **kwargs):
        self._json = self._UNDEFINED
        self._response_body = ''
        self._response_body_json = None

        RequestHandler.__init__(self, *args, **kwargs)

    def get_request_json(self):
        if self._json is self._UNDEFINED:
            try:
                self._json = json_utils.loads(self.request.body)

            except ValueError as e:
                logger.error('could not decode json from request body: %s', e)

                raise core_api.APIError(400, 'malformed body') from e

        return self._json

    def finish(self, chunk=None):
        self._response_body = chunk

        return super().finish(chunk)

    def finish_json(self, data):
        self._response_body_json = data

        data = json_utils.dumps(data)
        data += '\n'

        self.set_header('Content-Type', 'application/json; charset=utf-8')
        return self.finish(data)

    def get_response_body(self):
        return self._response_body

    def get_response_body_json(self):
        return self._response_body_json

    def get_response_headers(self):
        return dict(self._headers.get_all())

    def get(self, **kwargs):
        raise NoSuchFunction()

    head = post = delete = patch = put = options = get

    def _handle_request_exception(self, exception):
        try:
            if isinstance(exception, HTTPError):
                logger.error('%s %s: %s', self.request.method, self.request.uri, exception)
                self.set_status(exception.status_code)
                self.finish_json({'error': (exception.log_message or
                                            getattr(exception, 'reason', None) or str(exception))})

            else:
                logger.error(str(exception), exc_info=True)
                self.set_status(500)
                self.finish_json({'error': 'internal server error'})

        except RuntimeError:
            pass  # nevermind

    def data_received(self, chunk):
        pass


class StaticFileHandler(TornadoStaticFileHandler):
    def data_received(self, chunk):
        pass

    def set_extra_headers(self, path):
        self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate, max-age=0')


class JSModuleMapperStaticFileHandler(StaticFileHandler):
    def initialize(self, path, mapping, default_filename=None):
        super().initialize(path, default_filename)
        self._mapping = {k.encode(): v.encode() for k, v in mapping.items()}
        self._mapped_content = None

    def get_content_size(self):
        return len(self.get_mapped_content())

    def get_content(self, abspath, start=None, end=None):
        return self.get_mapped_content()

    @classmethod
    def get_content_version(cls, abspath):
        return ''

    def get_mapped_content(self):
        if self._mapped_content is None:
            content = b''.join(list(super().get_content(self.absolute_path)))

            if self.absolute_path.endswith('.js'):
                for k, v in self._mapping.items():
                    content = content.replace(k, v)

            self._mapped_content = content

        return self._mapped_content


class NoSuchFunctionHandler(BaseHandler):
    pass


class RedirectFrontendHandler(BaseHandler):
    def get(self):
        self.redirect(f'/{FRONTEND_URL_PREFIX}/')


class TemplateHandler(J2TemplateMixin, BaseHandler):
    def prepare(self):
        self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate, max-age=0')

    def get_context(self, path, offs=0):
        # Adjust static URL prefix to a relative path matching currently requested frontend path

        context = make_context()
        slashes = path.count('/') + offs
        if slashes == 0:
            prefix = 'frontend/'

        elif slashes > 1:
            prefix = '/'.join(['..'] * (slashes - 1)) + '/'

        else:
            prefix = ''

        context['static_url'] = prefix + context['static_url']

        return context


class FrontendHandler(TemplateHandler):
    def get(self, path):
        self.render('index.html', **self.get_context(path))


class ManifestHandler(TemplateHandler):
    def get(self, path):
        pretty_name = self.get_query_argument('pretty_name', None)
        description = self.get_query_argument('description', None)

        context = self.get_context(path, offs=1)
        if pretty_name:
            context['pretty_name'] = pretty_name
        if description:
            context['description'] = description

        self.set_header('Content-Type', 'application/manifest+json; charset="utf-8"')
        self.render('manifest.json', **context)


class ServiceWorkerHandler(TemplateHandler):
    def get(self):
        self.set_header('Content-Type', 'application/javascript; charset="utf-8"')
        self.render('service-worker.js', **make_context())


class APIHandler(BaseHandler):
    AUTH_ENABLED = True

    def __init__(self, *args, **kwargs):
        self.access_level = core_api.ACCESS_LEVEL_NONE
        self.username = None

        BaseHandler.__init__(self, *args, **kwargs)

    def prepare(self):
        # disable cache
        self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate, max-age=0')

        if not self.AUTH_ENABLED:
            return

        # parse auth header
        auth = self.request.headers.get('Authorization')
        if auth:
            try:
                usr = core_api_auth.parse_auth_header(auth, core_api_auth.ORIGIN_CONSUMER,
                                                      core_api_auth.consumer_password_hash_func)

            except core_api_auth.AuthError as e:
                logger.warning(str(e))
                return

        else:
            if core_device_attrs.admin_password_hash == core_device_attrs.EMPTY_PASSWORD_HASH:
                logger.debug('authenticating request as admin due to empty admin password')
                usr = 'admin'

            else:
                logger.warning('missing authorization header')
                return

        self.access_level = core_api.ACCESS_LEVEL_MAPPING[usr]
        self.username = usr

        logger.debug('granted access level %s (username=%s)',
                     core_api.ACCESS_LEVEL_MAPPING[self.access_level], self.username)

    async def call_api_func(self, func, default_status=200, **kwargs):
        try:
            if self.request.method in ('POST', 'PATCH', 'PUT'):
                kwargs['params'] = self.get_request_json()

            response = func(self, **kwargs)
            if inspect.isawaitable(response):
                response = await response

            self.set_status(default_status)
            if response is not None or default_status == 200:
                await self.finish_json(response)

            else:
                await self.finish()

        except Exception as e:
            await self._handle_api_call_exception(func, kwargs, e)

    async def _handle_api_call_exception(self, func, kwargs, error):
        kwargs = dict(kwargs)
        params = kwargs.pop('params', None)
        args = json_utils.dumps(kwargs)
        body = params and json_utils.dumps(params) or '{}'

        if isinstance(error, core_responses.HTTPError):
            error = core_api.APIError(error.code, error.msg)

        if isinstance(error, core_api.APIError):
            logger.error('api call %s failed: %s (args=%s, body=%s)', func.__name__, error, args, body)

            self.set_status(error.status)
            if not self._finished:  # Avoid finishing an already finished request
                await self.finish_json(dict({'error': error.message}, **error.params))

        elif isinstance(error, StreamClosedError) and func.__name__ == 'get_listen':
            logger.debug('api call get_listen could not complete: stream closed')

        else:
            logger.error('api call %s failed: %s (args=%s, body=%s)', func.__name__, error, args, body, exc_info=True)

            self.set_status(500)
            if not self._finished:  # Avoid finishing an already finished request
                await self.finish_json({'error': str(error)})


class DeviceHandler(APIHandler):
    async def get(self):
        await self.call_api_func(core_api_funcs.get_device)

    async def patch(self):
        await self.call_api_func(core_api_funcs.patch_device, default_status=204)


class ResetHandler(APIHandler):
    async def post(self):
        await self.call_api_func(core_api_funcs.post_reset, default_status=204)


class FirmwareHandler(APIHandler):
    async def get(self):
        await self.call_api_func(core_api_funcs.get_firmware)

    async def patch(self):
        await self.call_api_func(core_api_funcs.patch_firmware, default_status=204)


class AccessHandler(APIHandler):
    async def get(self):
        await self.call_api_func(core_api_funcs.get_access, access_level=self.access_level)


class SlaveDevicesHandler(APIHandler):
    async def get(self):
        await self.call_api_func(slaves_api_funcs.get_slave_devices)

    async def post(self):
        await self.call_api_func(slaves_api_funcs.post_slave_devices, default_status=201)


class SlaveDeviceHandler(APIHandler):
    async def patch(self, name):
        await self.call_api_func(slaves_api_funcs.patch_slave_device, name=name, default_status=204)

    async def delete(self, name):
        await self.call_api_func(slaves_api_funcs.delete_slave_device, name=name, default_status=204)


class SlaveDeviceEventsHandler(APIHandler):
    AUTH_ENABLED = False  # we'll take care of the authentication inside API call functions

    async def post(self, name):
        await self.call_api_func(slaves_api_funcs.post_slave_device_events, name=name, default_status=204)


class SlaveDeviceForwardHandler(APIHandler):
    async def get(self, name, path):
        await self.call_api_func(slaves_api_funcs.slave_device_forward,
                                 name=name, method=self.request.method, path=path)

    post = patch = delete = get


class PortsHandler(APIHandler):
    async def get(self):
        await self.call_api_func(core_api_funcs.get_ports)

    async def post(self):
        if not settings.core.virtual_ports:
            raise NoSuchFunction()

        await self.call_api_func(core_api_funcs.post_ports, default_status=201)


class PortHandler(APIHandler):
    async def delete(self, port_id):
        if not settings.core.virtual_ports:
            raise NoSuchFunction()

        await self.call_api_func(core_api_funcs.delete_port, port_id=port_id, default_status=204)

    async def patch(self, port_id):
        await self.call_api_func(core_api_funcs.patch_port, port_id=port_id, default_status=204)


class PortValueHandler(APIHandler):
    async def get(self, port_id):
        await self.call_api_func(core_api_funcs.get_port_value, port_id=port_id)

    async def patch(self, port_id):
        await self.call_api_func(core_api_funcs.patch_port_value, port_id=port_id, default_status=204)


class PortSequenceHandler(APIHandler):
    async def post(self, port_id):
        if not settings.core.sequences_support:
            raise NoSuchFunction()

        await self.call_api_func(core_api_funcs.post_port_sequence, port_id=port_id, default_status=204)


class WebhooksHandler(APIHandler):
    async def get(self):
        await self.call_api_func(core_api_funcs.get_webhooks)

    async def patch(self):
        await self.call_api_func(core_api_funcs.patch_webhooks, default_status=204)


class ListenHandler(APIHandler):
    async def get(self):
        session_id = self.get_argument('session_id', None)
        timeout = self.get_argument('timeout', None)

        await self.call_api_func(core_api_funcs.get_listen,
                                 session_id=session_id, timeout=timeout, access_level=self.access_level)


class ReverseHandler(APIHandler):
    async def get(self):
        await self.call_api_func(core_api_funcs.get_reverse)

    async def patch(self):
        await self.call_api_func(core_api_funcs.patch_reverse, default_status=204)


class DashboardPanelsHandler(APIHandler):
    async def get(self):
        await self.call_api_func(ui_api_funcs.get_panels)

    async def put(self):
        await self.call_api_func(ui_api_funcs.put_panels, default_status=204)


class PrefsHandler(APIHandler):
    async def get(self):
        await self.call_api_func(ui_api_funcs.get_prefs)

    async def put(self):
        await self.call_api_func(ui_api_funcs.put_prefs, default_status=204)




import logging

from qtoggleserver import persist


logger = logging.getLogger(__name__)


def reset():
    logger.debug('clearing slaves persisted data')
    persist.remove('slaves')
    persist.remove('slave_ports')




from .handlers import BaseEventHandler
from .handlers import handle_event
from .handlers import init as init_handlers

from .types.device import DeviceUpdate
from .types.port import PortAdd, PortRemove, PortUpdate, ValueChange
from .types.slave import SlaveDeviceAdd, SlaveDeviceRemove, SlaveDeviceUpdate


def init():
    init_handlers()




from qtoggleserver import persist
from qtoggleserver.core import api as core_api


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
def get_panels(request):
    return persist.get_value('dashboard_panels', default=[])


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
def put_panels(request, params):
    # core_api.validate(panels, PANELS_SCHEMA)  TODO validate panels against schema

    persist.set_value('dashboard_panels', params)


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
def get_prefs(request):
    # TODO add user id

    return persist.get_value('ui_prefs', {})


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
def put_prefs(request, params):
    # TODO add user id

    persist.set_value('ui_prefs', params)




from qtoggleserver import version
from qtoggleserver.conf import settings
from qtoggleserver.core.device import attrs as core_device_attrs

from . import constants


def make_context():
    return {
        'slaves_enabled': settings.slaves.enabled,

        'name': constants.APP_NAME,
        'pretty_name': constants.APP_PRETTY_NAME,
        'description': constants.APP_DESCRIPTION,
        'version': version.VERSION,
        'debug': settings.frontend.debug,
        'static_url': ['static', 'static/app'][settings.frontend.debug],
        'navigation_base_prefix': '/' + constants.FRONTEND_URL_PREFIX,
        'theme': core_device_attrs.ui_theme,
        'theme_color': constants.BROWSER_THEME_COLOR,
        'background_color': constants.BROWSER_BACKGROUND_COLOR
    }




import logging

from qtoggleserver import persist

from qtoggleserver.core import ports as core_ports


logger = logging.getLogger(__name__)


_vport_settings = {}


class VirtualPort(core_ports.Port):
    WRITABLE = True
    VIRTUAL = True

    def __init__(self, port_id, typ, mi, ma, integer, step, choices):
        super().__init__(port_id)

        self._type = typ
        self._min = mi
        self._max = ma
        self._integer = integer
        self._step = step
        self._choices = choices

        self._value = self._virtual_value = self.adapt_value_type_sync(typ, integer, mi or 0)

    def map_id(self, new_id):
        raise core_ports.PortError('virtual ports cannot be mapped')

    async def read_value(self):
        return self._virtual_value

    async def write_value(self, value):
        self._virtual_value = value


def add(port_id, typ, mi, ma, integer, step, choices):
    settings = {
        'id': port_id,
        'type': typ,
        'min': mi,
        'max': ma,
        'integer': integer,
        'step': step,
        'choices': choices
    }

    _vport_settings[port_id] = settings

    logger.debug('saving virtual port settings for %s', port_id)
    persist.replace('vports', port_id, settings)


def remove(port_id):
    try:
        _vport_settings.pop(port_id)

    except KeyError:
        logger.error('virtual port settings for %s no longer exist', port_id)

        return False

    logger.debug('removing virtual port settings for %s', port_id)
    persist.remove('vports', filt={'id': port_id})

    return True


def all_settings():
    return [dict({'driver': VirtualPort, 'port_id': port_id}, **settings)
            for port_id, settings in _vport_settings.items()]


def load():
    for entry in persist.query('vports'):
        _vport_settings[entry['id']] = {
            'typ': entry.get('type') or core_ports.TYPE_NUMBER,
            'mi': entry.get('min'),
            'ma': entry.get('max'),
            'integer': entry.get('integer'),
            'step': entry.get('step'),
            'choices': entry.get('choices')
        }

        logger.debug('loaded virtual port settings for %s', entry['id'])


def reset():
    logger.debug('clearing virtual ports persisted data')
    persist.remove('vports')




import abc
import asyncio
import logging
import os
import re

from qtoggleserver.lib import polled


W1_DEVICES_PATH = '/sys/bus/w1/devices'
SLAVE_FILE_NAME = 'w1_slave'


logger = logging.getLogger(__name__)


class OneWireException(Exception):
    pass


class OneWirePeripheralNotFound(OneWireException):
    def __init__(self, address):
        super().__init__('peripheral @{} not found'.format(address))


class OneWirePeripheralAddressRequired(OneWireException):
    def __init__(self):
        super().__init__('peripheral address required')


class OneWireTimeout(OneWireException):
    def __init__(self, message='timeout'):
        super().__init__(message)


class OneWirePeripheral(polled.PolledPeripheral):
    logger = logger

    TIMEOUT = 5  # Seconds

    def __init__(self, address, name):
        super().__init__(address, name)

        self._filename = None
        self._data = None
        self._error = None

    def get_filename(self):
        if self._filename is None:
            self._filename = self._find_filename()

        return self._filename

    def _find_filename(self):
        address_parts = re.split('[^a-zA-Z0-9]', self.get_address())
        pat = address_parts[0] + '-0*' + ''.join(address_parts[1:])
        for name in os.listdir(W1_DEVICES_PATH):
            if re.match(pat, name, re.IGNORECASE):
                return os.path.join(W1_DEVICES_PATH, name, SLAVE_FILE_NAME)

        raise OneWirePeripheralNotFound(self.get_address())

    def read(self):
        if self._error:
            error = self._error
            self._error = None

            raise error

        data = self._data
        self._data = None

        return data

    def read_sync(self):
        filename = self.get_filename()
        self.debug('opening file %s', filename)
        with open(filename, 'r') as f:
            data = f.read()
            self.debug('read data: %s', data.replace('\n', '\\n'))

        return data

    async def poll(self):
        try:
            future = self.run_threaded(self.read_sync)
            self._data = await asyncio.wait_for(future, timeout=self.TIMEOUT)

        except asyncio.TimeoutError:
            self._error = OneWireTimeout('timeout waiting for one-wire data from peripheral')

        except Exception as e:
            self._error = e

        else:
            self._error = None

    async def handle_disable(self):
        self._filename = None
        self._data = None
        self._error = None


class OneWirePort(polled.PolledPort, metaclass=abc.ABCMeta):
    PERIPHERAL_CLASS = OneWirePeripheral

    def __init__(self, address=None, peripheral_name=None):
        autodetected = False
        if address is None:
            address = self.autodetect_address()
            autodetected = True

        super().__init__(address, peripheral_name)

        if autodetected:
            self.debug('autodetected device address %s', address)

    @staticmethod
    def autodetect_address():
        # TODO make this method look only through specific device types (e.g. temperature sensors)

        names = os.listdir(W1_DEVICES_PATH)
        names = [n for n in names if re.match('^[0-9]{2}-', n)]
        if len(names) == 0:
            raise OneWirePeripheralNotFound('auto')

        if len(names) > 1:
            raise OneWirePeripheralAddressRequired()

        address = re.sub('[^a-f0-9]', '', names[0], re.IGNORECASE)
        address = ':'.join([address[2 * i: 2 * i + 2] for i in range(len(address) // 2)])

        return address




import errno
import socket


from qtoggleserver.utils import json as json_utils


class Error(Exception):
    MESSAGE = ''

    def __init__(self, **params):
        self._params = params

        super().__init__()

    def __str__(self):
        return self.MESSAGE.format(**self._params)

    def __repr__(self):
        return '{}("{}")'.format(self.__class__.__name__, self)


class HostUnreachable(Error):
    MESSAGE = 'host unreachable'


class NetworkUnreachable(Error):
    MESSAGE = 'network unreachable'


class UnresolvableHostname(Error):
    MESSAGE = 'hostname cannot be resolved'


class ConnectionRefused(Error):
    MESSAGE = 'connection refused'


class Timeout(Error):
    MESSAGE = 'timeout'


class MovedPermanently(Error):
    # HTTP 301
    MESSAGE = 'moved permanently to "{location}"'

    def __init__(self, location):
        self.location = location

        super().__init__(location=location)


class Redirect(Error):
    # HTTP 302, 303
    MESSAGE = 'redirected to "{location}"'

    def __init__(self, location):
        self.location = location

        super().__init__(location=location)


class HTTPError(Error):
    # 4xx - 5xx
    MESSAGE = '{code} {msg}'

    def __init__(self, code, msg):
        self.code = code
        self.msg = msg

        super().__init__(code=code, msg=msg)


class InvalidJson(Error):
    # JSON load() failure
    MESSAGE = 'invalid json'


class AuthError(Error):
    MESSAGE = 'authentication error: {msg}'

    def __init__(self, msg):
        super().__init__(msg=msg)


class OtherError(Error):
    # any other error
    MESSAGE = 'other error: {msg}'

    def __init__(self, msg):
        super().__init__(msg=msg)


def _response_error_errno(eno):
    if eno == errno.ECONNREFUSED:
        return ConnectionRefused()

    elif eno == errno.EHOSTUNREACH:
        return HostUnreachable()

    elif eno == errno.ENETUNREACH:
        return NetworkUnreachable()

    elif eno == socket.EAI_NONAME:
        return UnresolvableHostname()

    elif eno:
        return OtherError(errno.errorcode.get(eno))

    return OtherError('unknown error')


def parse(response, decode_json=True, resolve_refs=True):
    if 100 <= response.code < 599:
        if response.code == 204:
            return  # happy case - no content

        if decode_json and response.body:
            try:
                body = json_utils.loads(response.body, resolve_refs=resolve_refs)

            except Exception as e:
                raise InvalidJson() from e

        else:
            body = response.body

        if response.code == 200:
            return body  # happy case with content

        if response.code == 301:
            raise MovedPermanently(response.headers.get('Location', ''))

        if response.code in [302, 303]:
            raise Redirect(response.headers.get('Location', ''))

        if decode_json:
            raise HTTPError(response.code, body.get('error', ''))

        raise HTTPError(response.code, response.reason)

    elif response.error or response.code == 599:
        if str(response.error).lower().count('timeout'):
            raise Timeout()

        eno = getattr(response.error, 'errno', None)
        if eno:
            raise _response_error_errno(eno)

        raise OtherError(str(response.error))

    raise OtherError('unknown HTTP error ({}: {})'.format(response.code, str(response.error)))




import importlib
import logging
import os

from jinja2 import Environment, FileSystemLoader, select_autoescape

from .constants import FRONTEND_DIR


_env = None
logger = logging.getLogger(__name__)


class J2TemplateMixin:
    def render(self, template_name, **context):
        env = get_env()
        template = env.get_template(template_name)
        template_str = template.render(**context)

        self.finish(template_str)


class NamespaceLoader(FileSystemLoader):
    def __init__(self, namespace_name, path='templates', encoding='utf-8', followlinks=False):
        namespace = importlib.import_module(namespace_name)
        searchpath = [os.path.join(p, path) for p in namespace.__path__]

        super().__init__(searchpath=searchpath, encoding=encoding, followlinks=followlinks)


def get_env():
    global _env

    if _env is None:
        logger.debug('creating Jinja2 template environment')
        loader = NamespaceLoader('qtoggleserver', '{}/html'.format(FRONTEND_DIR))
        _env = Environment(loader=loader, autoescape=select_autoescape())

    return _env




import abc
import logging

from qtoggleserver import utils
from qtoggleserver.conf import settings


STATUS_IDLE = 'idle'
STATUS_CHECKING = 'checking'
STATUS_DOWNLOADING = 'downloading'
STATUS_EXTRACTING = 'extracting'
STATUS_VALIDATING = 'validating'
STATUS_FLASHING = 'flashing'
STATUS_RESTARTING = 'restarting'
STATUS_ERROR = 'error'


logger = logging.getLogger(__name__)

_fwupdate = None


class FWUpdateException(Exception):
    pass


class FWUpdateDisabled(FWUpdateException):
    pass


class BaseDriver(metaclass=abc.ABCMeta):
    def __init__(self):
        pass

    @abc.abstractmethod
    def get_current_version(self):
        pass

    @abc.abstractmethod
    def get_latest(self):
        pass

    @abc.abstractmethod
    def get_status(self):
        return STATUS_IDLE

    @abc.abstractmethod
    def update_to_version(self, version):
        pass

    @abc.abstractmethod
    def update_to_url(self, url):
        pass


def _get_fwupdate():
    global _fwupdate

    if not settings.system.fwupdate_driver:
        raise FWUpdateDisabled()

    if _fwupdate is None:
        logger.debug('initializing fwupdate')

        cls = utils.load_attr(settings.system.fwupdate_driver)
        _fwupdate = cls()

    return _fwupdate


async def get_current_version():
    current_version = await _get_fwupdate().get_current_version()
    logger.debug('current version: %s', current_version)

    return current_version


async def get_latest():
    latest_version, latest_date, latest_url = await _get_fwupdate().get_latest()
    logger.debug('latest version: %s/%s at %s', latest_version, latest_date, latest_url)

    return latest_version, latest_date, latest_url


async def get_status():
    status = await _get_fwupdate().get_status()
    logger.debug('status: %s', status)

    return status


async def update_to_version(version):
    logger.debug('updating to version %s', version)

    await _get_fwupdate().update_to_version(version)


async def update_to_url(url):
    logger.debug('updating to url %s', url)

    await _get_fwupdate().update_to_url(url)




from qtoggleserver.core import ports
from qtoggleserver.utils import json as json_utils


class DummyGPIO(ports.Port):
    TYPE = ports.TYPE_BOOLEAN

    ADDITIONAL_ATTRDEFS = {
        'output': {
            'display_name': 'Is Output',
            'description': 'Controls the port direction.',
            'type': 'boolean',
            'modifiable': True
        }
    }

    def __init__(self, no, def_value=None, def_output=None):
        self._no = no

        self._def_value = def_value
        self._def_output = def_output

        self._dummy_value = def_value
        self._dummy_output = def_output if def_output is not None else False

        super().__init__(port_id='gpio{}'.format(no))

    async def handle_enable(self):
        if self._def_output is not None:
            await self.attr_set_output(self._def_output)

    async def read_value(self):
        return self._dummy_value

    async def write_value(self, value):
        self.debug('writing "%s"', json_utils.dumps(value))
        self._dummy_value = value

    async def attr_is_writable(self):
        return self._dummy_output

    async def attr_set_output(self, output):
        self._dummy_output = output

        if output:
            self.debug('setting output mode')

        else:
            self.debug('setting input mode')

        if output and self._def_value is not None:
            await self.write_value(self._def_value)

    async def attr_is_output(self):
        return self._dummy_output




import functools
import inspect
import logging


logger = logging.getLogger(__name__)

_init_hooks = []
_cleanup_hooks = []


def add_init_hook(hook, *args, **kwargs):
    _init_hooks.append(functools.partial(hook, *args, **kwargs))


def add_cleanup_hook(hook, *args, **kwargs):
    _cleanup_hooks.append(functools.partial(hook, *args, **kwargs))


async def init():
    for hook in _init_hooks:
        try:
            result = hook()
            if inspect.isawaitable(result):
                await result

        except Exception as e:
            logger.error('init hook failed: %s', e, exc_info=True)


async def cleanup():
    for hook in _cleanup_hooks:
        try:
            result = hook()
            if inspect.isawaitable(result):
                await result

        except Exception as e:
            logger.error('cleanup hook failed: %s', e, exc_info=True)




import os as _os


debug = False


logging = {
    'version': 1,
    'memory_logs_buffer_len': 10000,
    'formatters': {
        'default': {
            'format': '%(asctime)s: %(levelname)7s: [%(name)s] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'default'
        }
    },
    'loggers': {
        '"qtoggleserver.persist"': {'level': 'INFO'}  # Double quotes are necessary to avoid HOCON key split
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console']
    }
}


class core:
    tick_interval = 50
    event_queue_size = 256
    max_client_time_skew = 300
    listen_support = True
    sequences_support = True
    ssl_support = False
    virtual_ports = 1024


class server:
    addr = '0.0.0.0'
    port = 8888
    compress_response = False

    class https:
        cert_file = None
        key_file = None


class persist:
    driver = 'qtoggleserver.drivers.persist.redis.RedisDriver'
    host = '127.0.0.1'
    port = 6379
    db = 0


class system:
    class date:
        set_cmd = None
        set_format = '%Y-%m-%dT%H:%M:%SZ'

    class timezone:
        get_cmd = None
        set_cmd = None

    class net:
        class wifi:
            get_cmd = None
            set_cmd = None

        class ip:
            get_cmd = None
            set_cmd = None

    fwupdate_driver = None


class frontend:
    enabled = True
    debug = False


class slaves:
    enabled = False
    timeout = 10
    keepalive = 300
    retry_interval = 5
    retry_count = 3


class webhooks:
    enabled = False


class reverse:
    enabled = False
    retry_interval = 5


class configurables:
    pass


class device_name:
    get_cmd = None
    set_cmd = None


password_set_cmd = None

event_handlers = []

ports = []

port_mappings = {}

pkg_path = _os.path.dirname(_os.path.dirname(_os.path.abspath(__file__)))




import abc
import functools
import logging
import re
import subprocess
import time

from bluepy import btle

from qtoggleserver import utils
from qtoggleserver.core import ports as core_ports

from . import polled
from .peripheral import Peripheral, RunnerBusy


logger = logging.getLogger(__name__)


class BLEException(Exception):
    pass


class BLEBusy(BLEException):
    def __init__(self, message='busy'):
        super().__init__(message)


class BLETimeout(BLEException):
    def __init__(self, message='timeout'):
        super().__init__(message)


class _BluepyTimeoutError(btle.BTLEException):
    def __init__(self, message, rsp=None):
        super().__init__(message, rsp)


class _BluepyPeripheral(btle.Peripheral):
    def __init__(self, timeout=None, *args, **kwargs):
        self._timeout = timeout
        super().__init__(*args, **kwargs)

    def _getResp(self, wantType, timeout=None):
        # Override this method to be able to inject a default timeout
        # We also need to raise a timeout exception in case the response is None

        timeout = timeout or self._timeout
        response = super()._getResp(wantType, timeout)
        if response is None:
            raise _BluepyTimeoutError('timeout waiting for a response from peripheral')

        return response

    def _stopHelper(self):
        # Override this method to close the helper's stdout and stdin streams.
        helper = self._helper
        super()._stopHelper()
        if helper:
            helper.stdin.close()
            helper.stdout.close()


class BLEAdapter(utils.ConfigurableMixin, utils.LoggableMixin):
    DEFAULT_NAME = 'hci0'
    RUNNER_CLASS = Peripheral.RUNNER_CLASS

    _adapters_by_name = {}

    @classmethod
    def get(cls, name):
        if name not in cls._adapters_by_name:
            logger.debug('initializing adapter %s', name)
            adapter = cls(name)
            cls._adapters_by_name[name] = adapter

        return cls._adapters_by_name[name]

    def __init__(self, name):
        utils.LoggableMixin.__init__(self, name, logger)

        self._name = name
        try:
            self._address = self._find_address(name)

        except Exception as e:
            self.error('could not initialize adapter: %s', e)
            raise

        self.debug('found adapter address %s', self._address)

        self._runner = None
        self._peripherals = []

    def __str__(self):
        return self._name

    def add_peripheral(self, peripheral):
        self._peripherals.append(peripheral)

    def get_runner(self):
        if self._runner is None:
            self._runner = self.make_runner()

        return self._runner

    def make_runner(self):
        self.debug('starting threaded runner')
        runner = self.RUNNER_CLASS(queue_size=Peripheral.RUNNER_QUEUE_SIZE)
        runner.start()

        return runner

    @staticmethod
    def pretty_data(data):
        return ' '.join(map(lambda c: '{:02X}'.format(c), data))

    @staticmethod
    def _find_address(name):
        try:
            output = subprocess.check_output(['hcitool', '-i', name, 'dev'], stderr=subprocess.STDOUT).decode()

        except subprocess.CalledProcessError as e:
            output = e.output.decode()
            if output.count('No such device'):
                raise Exception('adapter not found: {}'.format(name)) from e

            else:
                raise Exception(output.strip()) from e

        found = re.findall(name + r'\s([0-9a-f:]{17})', output, re.IGNORECASE)
        if not found:
            raise Exception('adapter not found: {}'.format(name))

        return found[0]


class BLEPeripheral(polled.PolledPeripheral, metaclass=abc.ABCMeta):
    CMD_TIMEOUT = 20
    RETRY_COUNT = 3
    WRITE_VALUE_PAUSE = 5

    logger = logger

    @classmethod
    def make_peripheral(cls, address, name, adapter_name=None, **kwargs):
        return cls(address, name, BLEAdapter.get(adapter_name))

    def __init__(self, address, name, adapter):
        super().__init__(address, name)

        self._adapter = adapter
        self._adapter.add_peripheral(self)
        self._online = False

    def __str__(self):
        return '{}/{}'.format(self._adapter, self._name)

    def make_runner(self):
        # Instead of creating a runner for each peripheral instance, we use adapter's runner, thus having one runner per
        # adapter instance.

        return self._adapter.get_runner()

    async def read(self, handle, retry_count=None):
        if retry_count is None:
            retry_count = self.RETRY_COUNT

        return await self._run_cmd_async('read', self.get_address(), handle, notify_handle=None,
                                         data=None, timeout=self.CMD_TIMEOUT, retry_count=retry_count)

    async def write(self, handle, data, retry_count=None):
        if retry_count is None:
            retry_count = self.RETRY_COUNT

        return await self._run_cmd_async('write', self.get_address(), handle, notify_handle=None,
                                         data=data, timeout=self.CMD_TIMEOUT, retry_count=retry_count)

    async def write_notify(self, handle, notify_handle, data, timeout=None, retry_count=None):
        if retry_count is None:
            retry_count = self.RETRY_COUNT

        if timeout is None:
            timeout = self.CMD_TIMEOUT

        return await self._run_cmd_async('write', self.get_address(), handle, notify_handle=notify_handle,
                                        data=data, timeout=timeout, retry_count=retry_count)

    def _run_cmd(self, cmd, address, handle, notify_handle, data, timeout):
        start_time = time.time()
        response = None
        notification_data = None

        self.debug('connecting to %s', address)

        bluepy_peripheral = _BluepyPeripheral(timeout=timeout)
        bluepy_peripheral.connect(address)

        if notify_handle:
            # Create a temporary class to deal with BTLE delegation

            peripheral = self
            class Delegate(btle.DefaultDelegate):

                def handleNotification(self, h, d):
                    nonlocal notification_data

                    if not notify_handle or h != notify_handle:
                        peripheral.warning('got unexpected notification on %04X', h)
                        return

                    peripheral.debug('got notification on %04X: %s', h, BLEPeripheral.pretty_data(d))
                    notification_data = d

            bluepy_peripheral.withDelegate(Delegate())

        if cmd == 'write':
            self.debug('writing at %04X: %s', handle, self.pretty_data(data))
            bluepy_peripheral.writeCharacteristic(handle, data, withResponse=True)

        else:  # assuming read
            self.debug('reading from %04X', handle)
            response = bluepy_peripheral.readCharacteristic(handle)
            self.debug('got response %s', self.pretty_data(response))

        if notify_handle and notification_data is None:
            self.debug('waiting for notification on %04X', notify_handle)

            while not notification_data and self.get_runner().is_running():
                if time.time() - start_time > timeout:
                    raise BLETimeout('timeout waiting for notification on {:04X}'.format(notify_handle))

                try:
                    bluepy_peripheral.waitForNotifications(0.1)

                except _BluepyTimeoutError:
                    continue

        return response, notification_data

    async def _run_cmd_async(self, cmd, address, handle, notify_handle, data, timeout, retry_count):
        retry = 1
        while True:
            try:
                response, notification_data = await self.run_threaded(self._run_cmd, cmd, address, handle,
                                                                      notify_handle, data, timeout)

            except Exception as e:
                self.error('command execution failed: %s', e)

                if retry <= retry_count:
                    self.warning('retry %s/%s', retry, retry_count)
                    retry += 1
                    continue

                if self._online:
                    self._online = False
                    self._handle_offline()

                if isinstance(e, RunnerBusy):
                    raise BLEBusy('too many pending commands') from e

                raise

            else:
                if not self._online:
                    self._online = True
                    self._handle_online()

                return response, notification_data

    def is_online(self):
        return self._enabled and self._online

    def _handle_offline(self):
        self.debug('%s is offline', self)
        self.trigger_port_update()

    def _handle_online(self):
        self.debug('%s is online', self)
        self.trigger_port_update()

    @staticmethod
    def pretty_data(data):
        return BLEAdapter.pretty_data(data)


class BLEPort(polled.PolledPort, metaclass=abc.ABCMeta):
    PERIPHERAL_CLASS = BLEPeripheral

    READ_INTERVAL_MAX = 1440
    READ_INTERVAL_STEP = 1
    READ_INTERVAL_MULTIPLIER = 60

    def __init__(self, address, peripheral_name=None, adapter_name=None):
        if adapter_name is None:
            adapter_name = BLEAdapter.DEFAULT_NAME

        super().__init__(address, peripheral_name, adapter_name=adapter_name)

    async def attr_get_write_value_pause(self):
        # Inherit from peripheral
        return self.get_peripheral().WRITE_VALUE_PAUSE

    async def attr_is_online(self):
        if not self.is_enabled():
            return False

        return self.get_peripheral().is_online()


def port_exceptions(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # transform BLE exceptions into port exceptions, where applicable
        try:
            return func(*args, **kwargs)

        except BLETimeout as e:
            raise core_ports.PortTimeout() from e

        except BLEException as e:
            raise core_ports.PortError(str(e)) from e

    return wrapper




import argparse
import asyncio
import logging.config
import signal
import sys

from tornado import httpclient

from qtoggleserver import lib
from qtoggleserver import persist
from qtoggleserver import utils
from qtoggleserver import version
from qtoggleserver.conf import settings
from qtoggleserver.conf import utils as conf_utils
from qtoggleserver.core import device
from qtoggleserver.core import events
from qtoggleserver.core import main
from qtoggleserver.core import ports
from qtoggleserver.core import reverse
from qtoggleserver.core import vports
from qtoggleserver.core import webhooks
from qtoggleserver.slaves import devices as slaves_devices
from qtoggleserver.utils.misc import FifoMemoryHandler


logger = None
options = None

_stopping = False


def parse_args():
    global options

    description = 'qToggleServer {}'.format(version.VERSION)
    epilog = None

    class VersionAction(argparse.Action):
        def __call__(self, *args, **kwargs):
            sys.stdout.write('{}\n'.format(description))
            sys.exit()

    parser = argparse.ArgumentParser(prog=sys.argv[0],
                                     description=description, epilog=epilog,
                                     add_help=False, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-c', help='specify the configuration file',
                        type=str, dest='config_file')
    parser.add_argument('-h', help='print this help and exit',
                        action='help', default=argparse.SUPPRESS)
    parser.add_argument('-v', help='print program version and exit',
                        action=VersionAction, default=argparse.SUPPRESS, nargs=0)

    options = parser.parse_args()

    if not options.config_file:
        parser.print_usage()
        sys.exit(-1)


def init_settings():
    try:
        parsed_config = conf_utils.config_from_file(options.config_file)

    except IOError as e:
        sys.stderr.write('failed to open config file "{}": {}\n'.format(options.config_file, e))
        sys.exit(-1)

    except Exception as e:
        sys.stderr.write('failed to load config file "{}": {}\n'.format(options.config_file, e))
        sys.exit(-1)

    def_config = conf_utils.obj_to_dict(settings)
    def_config = conf_utils.config_from_dict(def_config)

    config = conf_utils.config_merge(def_config, parsed_config)
    config = conf_utils.config_to_dict(config)

    conf_utils.update_obj_from_dict(settings, config)


def init_logging():
    global logger

    settings.logging['disable_existing_loggers'] = False
    logging.config.dictConfig(settings.logging)

    # add memory logs handler
    root_logger = logging.getLogger()
    main.memory_logs = FifoMemoryHandler(capacity=settings.logging['memory_logs_buffer_len'])
    root_logger.addHandler(main.memory_logs)

    logger = logging.getLogger('qtoggleserver')

    logger.info('hello!')
    logger.info('this is qToggleServer %s', version.VERSION)

    # we can't do this in init_settings() because we have no logging there
    logger.info('using config from %s', options.config_file)


def init_signals():
    loop = asyncio.get_event_loop()

    def bye_handler(sig, frame):
        global _stopping

        if _stopping:
            logger.error('interrupt signal already received, ignoring')
            return

        _stopping = True

        logger.info('interrupt signal received, shutting down')

        loop.call_soon_threadsafe(loop.stop)

    signal.signal(signal.SIGINT, bye_handler)
    signal.signal(signal.SIGTERM, bye_handler)


def init_configurables():
    httpclient.AsyncHTTPClient.configure('tornado.simple_httpclient.SimpleAsyncHTTPClient', max_clients=1024)

    configurables = conf_utils.obj_to_dict(settings.configurables)
    for class_name, opts in sorted(configurables.items()):
        try:
            logger.debug('configuring class %s', class_name)
            klass = utils.load_attr(class_name)

        except Exception as e:
            logger.error('failed to load class %s: %s', class_name, e, exc_info=True)
            continue

        try:
            klass.configure(**opts)

        except Exception as e:
            logger.error('failed to configure class %s: %s', class_name, e, exc_info=True)


def init_persist():
    logger.info('initializing persistence')
    try:
        persist.get_value('device')

    except Exception as e:
        logger.error('failed to initialize persistence: %s', e, exc_info=True)
        sys.exit(-1)


async def cleanup_persist():
    persist.close()


def init_events():
    logger.info('initializing events')
    events.init()


async def init_device():
    logger.info('initializing device')
    device.load()

    if settings.webhooks.enabled:
        logger.info('initializing webhooks')
        webhooks.load()

    if settings.reverse.enabled:
        logger.info('initializing reverse API calls')
        reverse.load()


async def init_ports():
    logger.info('initializing ports')
    vports.load()
    port_settings = settings.ports + vports.all_settings()
    await ports.load(port_settings)


async def cleanup_ports():
    logger.info('cleaning up ports')
    await ports.cleanup()


async def init_slaves():
    if settings.slaves.enabled:
        logger.info('initializing slaves')
        slaves_devices.load()


async def cleanup_slaves():
    if settings.slaves.enabled:
        logger.info('cleaning up slaves')
        await slaves_devices.cleanup()


async def init_lib():
    logger.info('initializing libs')
    await lib.init()


async def cleanup_lib():
    logger.info('cleaning up libs')
    await lib.cleanup()


async def init_main():
    logger.info('initializing main')
    await main.init()

    # Wait until slaves are also ready before actually considering main loop ready
    if settings.slaves.enabled:
        logger.info('waiting for slaves to become ready')
        while not slaves_devices.ready():
            await asyncio.sleep(0.1)

    # Mark main as ready after all slaves with their ports have been initialized and hopefully brought online. Allow an
    # extra second for pending loop tasks.
    await asyncio.sleep(1)
    main.set_ready()


async def cleanup_main():
    logger.info('cleaning up main')
    await main.cleanup()


async def init():
    parse_args()

    init_settings()
    init_logging()
    init_signals()
    init_configurables()
    init_persist()
    init_events()

    await init_device()
    await init_ports()
    await init_slaves()
    await init_main()
    await init_lib()


async def cleanup():
    await cleanup_slaves()
    await cleanup_ports()
    await cleanup_lib()
    await cleanup_main()
    await cleanup_persist()




import pytz
import time

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd, run_set_cmd


OLD_TIME_LIMIT = 1546304400  # January 2019


class DateError(Exception):
    pass


def has_real_date_time():
    # This is to distinguish between unset date (normally around 1970) and a system that has real date time support
    return time.time() > OLD_TIME_LIMIT


def has_date_support():
    return bool(settings.system.date.set_cmd)


def set_date(date):
    date_str = date.strftime(settings.system.date.set_format)
    run_set_cmd(settings.system.date.set_cmd, cmd_name='date', exc_class=DateError, date=date_str)


def has_timezone_support():
    return bool(settings.system.timezone.get_cmd and settings.system.timezone.set_cmd)


def get_timezone():
    return run_get_cmd(settings.system.timezone.get_cmd, cmd_name='timezone',
                       exc_class=DateError, required_fields=['timezone'])['timezone']


def set_timezone(timezone):
    run_set_cmd(settings.system.timezone.set_cmd, cmd_name='timezone', exc_class=DateError, timezone=timezone)


def get_timezones():
    return pytz.common_timezones




import asyncio
import logging


logger = logging.getLogger(__name__)


class Sequence:
    def __init__(self, values, delays, repeat, callback, finish_callback):
        self._values = values
        self._delays = delays
        self._repeat = repeat

        self._callback = callback
        self._finish_callback = finish_callback
        self._cancelled = False
        self._counter = 0

    def start(self):
        asyncio.create_task(self._loop())

    def cancel(self):
        self._cancelled = True

    async def _loop(self):
        for i, value in enumerate(self._values):
            if self._cancelled:
                break

            try:
                self._callback(value)

            except Exception as e:
                logger.error('sequence callback failed: %s', e, exc_info=True)

            if i < len(self._values) - 1:
                await asyncio.sleep(self._delays[i] / 1000.0)

            else:
                if self._repeat > 0 and self._counter >= self._repeat - 1:
                    await self._finish_callback()

                    return

                self._counter += 1
                await asyncio.sleep(self._delays[i] / 1000.0)

                asyncio.create_task(self._loop())




from qtoggleserver.core import api as core_api

from .base import Event


class PortEvent(Event):
    def __init__(self, port, timestamp=None):
        self._port = port

        super().__init__(timestamp)

    def __str__(self):
        return '{}({}) event'.format(self._type, self._port.get_id())

    def get_port(self):
        return self._port


class PortAdd(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'port-add'

    async def get_params(self):
        return await self.get_port().to_json()


class PortRemove(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'port-remove'

    async def get_params(self):
        return {'id': self.get_port().get_id()}


class PortUpdate(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'port-update'

    async def get_params(self):
        return await self.get_port().to_json()

    def is_duplicate(self, event):
        return isinstance(event, self.__class__) and event.get_port() == self.get_port()


class ValueChange(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'value-change'

    async def get_params(self):
        return {'id': self.get_port().get_id(), 'value': self.get_port().get_value()}

    def is_duplicate(self, event):
        return isinstance(event, self.__class__) and event.get_port() == self.get_port()




import asyncio

import time
import logging

from qtoggleserver.conf import settings


logger = logging.getLogger(__name__)

_sessions_by_id = {}


class Session:
    def __init__(self, session_id):
        self.id = session_id
        self.accessed = 0
        self.timeout = 0
        self.access_level = 0
        self.future = None
        self.queue = []

    def reset_and_wait(self, timeout, access_level):
        logger.debug('resetting %s (timeout=%s, access_level=%s)', self, timeout, access_level)

        if self.future:
            logger.debug('%s already has a listening connection, responding', self)
            self.respond()

        future = asyncio.get_running_loop().create_future()

        self.accessed = time.time()
        self.timeout = timeout
        self.access_level = access_level
        self.future = future

        if self.queue:
            logger.debug('%s has queued events, responding right away', self)
            self.respond()

        return future

    def is_empty(self):
        return len(self.queue) == 0

    def is_active(self):
        return self.future is not None

    def respond(self):
        events = list(self.queue)
        self.queue = []
        if not self.future:
            return

        self.future.set_result(events)
        self.future = None

    def push(self, event):
        # Deduplicate events
        while True:
            duplicates = [e for e in self.queue if event.is_duplicate(e)]
            if not duplicates:
                break

            for d in duplicates:
                self.queue.remove(d)
                logger.debug('dropping duplicate event %s from %s', d, self)

        # Ensure max queue size
        while len(self.queue) >= settings.core.event_queue_size:
            logger.debug('%s queue full, dropping oldest event', self)
            self.queue.pop()

        self.queue.insert(0, event)

    def __str__(self):
        return 'session {}'.format(self.id)


def get(session_id):
    session = _sessions_by_id.get(session_id)
    if not session:
        session = Session(session_id)
        _sessions_by_id[session_id] = session
        logger.debug('%s created', session)

    return session


def push(event):
    logger.debug('%s triggered', event)

    for session in _sessions_by_id.values():
        if session.access_level < event.REQUIRED_ACCESS:
            continue

        session.push(event)


def respond_non_empty():
    for session in _sessions_by_id.values():
        if session.is_empty():
            continue

        if not session.is_active():
            continue

        session.respond()


def cleanup():
    now = time.time()
    for session_id, session in list(_sessions_by_id.items()):
        if now - session.accessed > session.timeout:
            if session.is_active():
                logger.debug('%s keep-alive', session)
                session.respond()

            else:
                logger.debug('%s expired', session)
                _sessions_by_id.pop(session_id)




import functools
import logging


API_VERSION = '1.0'

ACCESS_LEVEL_ADMIN = 30
ACCESS_LEVEL_NORMAL = 20
ACCESS_LEVEL_VIEWONLY = 10
ACCESS_LEVEL_NONE = 0

ACCESS_LEVEL_MAPPING = {
    ACCESS_LEVEL_ADMIN: 'admin',
    ACCESS_LEVEL_NORMAL: 'normal',
    ACCESS_LEVEL_VIEWONLY: 'viewonly',
    ACCESS_LEVEL_NONE: 'none',
    'admin': ACCESS_LEVEL_ADMIN,
    'normal': ACCESS_LEVEL_NORMAL,
    'viewonly': ACCESS_LEVEL_VIEWONLY,
    'none': ACCESS_LEVEL_NONE
}

logger = logging.getLogger(__name__)


class APIError(Exception):
    def __init__(self, status, message, **params):
        self.status = status
        self.message = message
        self.params = params

        super().__init__(message)

    @staticmethod
    def from_http_error(http_error):
        return APIError(http_error.code, http_error.msg)


class APIRequest:
    def __init__(self, handler):
        self.handler = handler

    @property
    def access_level(self):
        return self.handler.access_level

    @property
    def method(self):
        return self.handler.request.method

    @property
    def path(self):
        return self.handler.request.path

    @property
    def query_arguments(self):
        return {k: self.handler.decode_argument(v[0]) for k, v in self.handler.request.query_arguments.items()}

    @property
    def headers(self):
        return self.handler.request.headers

    @property
    def body(self):
        return self.handler.request.body


def api_call(access_level=ACCESS_LEVEL_NONE):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(request_handler, *args, **kwargs):
            logger.debug('executing API call "%s"', func.__name__)

            if request_handler.access_level < access_level:
                if request_handler.access_level == ACCESS_LEVEL_NONE:  # indicates missing or invalid auth data
                    raise APIError(401, 'authentication required')

                else:
                    raise APIError(403, 'forbidden', required_level=ACCESS_LEVEL_MAPPING.get(access_level))

            request = APIRequest(request_handler)

            return func(request, *args, **kwargs)

        return wrapper

    return decorator




from qtoggleserver.core import api as core_api

from .base import Event


class SlaveDeviceEvent(Event):
    def __init__(self, slave, timestamp=None):
        self._slave = slave

        super().__init__(timestamp)

    def __str__(self):
        return '{}({}) event'.format(self._type, self._slave.get_name())

    def get_slave(self):
        return self._slave


class SlaveDeviceAdd(SlaveDeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'slave-device-add'

    async def get_params(self):
        return self.get_slave().to_json()


class SlaveDeviceRemove(SlaveDeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'slave-device-remove'

    async def get_params(self):
        return {'name': self.get_slave().get_name()}


class SlaveDeviceUpdate(SlaveDeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'slave-device-update'

    async def get_params(self):
        return self.get_slave().to_json()

    def is_duplicate(self, event):
        return isinstance(event, self.__class__) and event.get_slave() == self.get_slave()




import logging
import os

from qtoggleserver.conf import settings

from . import date
from . import fwupdate
from . import net


logger = logging.getLogger(__name__)


def reboot():
    logger.debug('rebooting')
    if not settings.debug:
        os.system('reboot')


def uptime():
    with open('/proc/uptime', 'r') as f:
        line = f.readlines()[0]
        return int(float(line.split()[0]))




FRONTEND_DIR = 'frontend'
FRONTEND_URL_PREFIX = 'frontend'

APP_NAME = 'qtoggleserver'
APP_PRETTY_NAME = 'qToggle Remote'
APP_DESCRIPTION = 'Control your qToggle devices'

BROWSER_BACKGROUND_COLOR = '#444444'
BROWSER_THEME_COLOR = '#62abea'




import asyncio
import logging
import sys


# TODO move these to modules

class ConfigurableMixin:
    @classmethod
    def configure(cls, **kwargs):
        for name, value in kwargs.items():
            conf_method = getattr(cls, 'configure_{}'.format(name.lower()), None)
            if conf_method:
                conf_method(value)

            else:
                setattr(cls, name.upper(), value)


class LoggableMixin:
    def __init__(self, name, parent_logger=None):
        if parent_logger and name:
            name = '{}.{}'.format(parent_logger.name, name)

        else:
            name = parent_logger.name

        self._parent_logger = parent_logger
        self._logger = logging.getLogger(name)

    def log(self, level, msg, *args, **kwargs):
        self._logger.log(level, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

    def set_logger_name(self, name):
        if name:
            self._logger = logging.getLogger('{}.{}'.format(self._parent_logger.name, name))

        else:
            self._logger = self._parent_logger


def load_attr(attr_path):
    m, attr = attr_path.rsplit('.', 1)

    try:
        __import__(m)
        mod = sys.modules[m]

    except ImportError as e:
        raise Exception('Error importing {}: {}'.format(attr_path, e)) from e

    try:
        attr = getattr(mod, attr)

    except AttributeError as e:
        raise Exception('Error importing {}: {}'.format(attr_path, e)) from e

    return attr


async def await_later(delay, coro, *args, loop=None):
    await asyncio.sleep(delay, loop=loop)
    await coro(*args)




from qtoggleserver.core import events as core_events
from qtoggleserver.core import sessions as core_sessions


def trigger_update():
    event = core_events.DeviceUpdate()
    core_sessions.push(event)
    core_events.handle_event(event)




import logging
import os.path

from tornado.web import Application

from qtoggleserver.conf import settings
from qtoggleserver.web import handlers

from .constants import FRONTEND_DIR, FRONTEND_URL_PREFIX


logger = logging.getLogger(__name__)

_application = None


def _log_request(handler):
    if handler.get_status() < 400:
        log_method = logger.debug

    elif handler.get_status() < 500:
        log_method = logger.warning

    else:
        log_method = logger.error

    request_time = 1000.0 * handler.request.request_time()
    # noinspection PyProtectedMember
    log_method('%d %s %.2fms', handler.get_status(), handler._request_summary(), request_time)


def _make_handlers():
    handlers_list = []

    # frontend

    if settings.frontend.enabled:
        frontend_path = os.path.join(settings.pkg_path, FRONTEND_DIR)
        if not os.path.exists(frontend_path):  # Look outside of main package (dev mode)
            frontend_path = os.path.join(settings.pkg_path, '..', FRONTEND_DIR)

        if settings.frontend.debug:
            js_module_path_mapping = {
                '$qui': f'/{FRONTEND_URL_PREFIX}/static/qui/js',
                '$app': f'/{FRONTEND_URL_PREFIX}/static/app/js'
            }

            # In debug mode, we serve QUI static files from its own folder, assumed to be in node_modules/@qtoggle
            qui_path = os.path.join(frontend_path, 'node_modules', '@qtoggle', 'qui')

            handlers_list.append((fr'^/{FRONTEND_URL_PREFIX}/static/qui/(.*)$',
                                  handlers.JSModuleMapperStaticFileHandler,
                                  {'path': qui_path, 'mapping': js_module_path_mapping}))

            handlers_list.append((fr'^/{FRONTEND_URL_PREFIX}/static/app/(.*)$',
                                  handlers.JSModuleMapperStaticFileHandler,
                                  {'path': frontend_path, 'mapping': js_module_path_mapping}))

        else:
            if os.path.exists(os.path.join(frontend_path, 'dist')):  # "dist" folder (prod mode, unpackaged)
                frontend_path = os.path.join(frontend_path, 'dist')

            handlers_list.append((fr'^/{FRONTEND_URL_PREFIX}/static/(.*)$',
                                  handlers.StaticFileHandler,
                                  {'path': frontend_path}))

        handlers_list += [
            (r'^/?$', handlers.RedirectFrontendHandler),
            (fr'^/{FRONTEND_URL_PREFIX}/service-worker.js$', handlers.ServiceWorkerHandler),
            (fr'^/{FRONTEND_URL_PREFIX}(?P<path>.*)/manifest.json$', handlers.ManifestHandler),
            (fr'^/{FRONTEND_URL_PREFIX}(?P<path>.*)', handlers.FrontendHandler),
        ]

        handlers_list += [
            (fr'^/api/frontend/dashboard/panels/?$', handlers.DashboardPanelsHandler),
            (fr'^/api/frontend/prefs/?$', handlers.PrefsHandler),
        ]

    handlers_list += [
        # device management
        (r'^/api/device/?$', handlers.DeviceHandler),
        (r'^/api/reset/?$', handlers.ResetHandler),
        (r'^/api/access/?$', handlers.AccessHandler),

        # port management
        (r'^/api/ports/?$', handlers.PortsHandler),
        (r'^/api/ports/(?P<port_id>[\w.]+)/?$', handlers.PortHandler),

        # port values
        (r'^/api/ports/(?P<port_id>[\w.]+)/value/?$', handlers.PortValueHandler),
    ]

    if settings.core.sequences_support:
        handlers_list += [
            (r'^/api/ports/(?P<port_id>[\w.]+)/sequence/?$', handlers.PortSequenceHandler)
        ]

    # firmware

    if settings.system.fwupdate_driver:
        handlers_list += [
            (r'^/api/firmware/?$', handlers.FirmwareHandler)
        ]

    # slave devices management

    if settings.slaves.enabled:
        handlers_list += [
            (r'^/api/devices/?$', handlers.SlaveDevicesHandler),
            (r'^/api/devices/(?P<name>\w+)/?$', handlers.SlaveDeviceHandler),
            (r'^/api/devices/(?P<name>\w+)/events/?$', handlers.SlaveDeviceEventsHandler),
            (r'^/api/devices/(?P<name>\w+)/forward/(?P<path>.+)$', handlers.SlaveDeviceForwardHandler)
        ]

    # notifications

    if settings.webhooks.enabled:
        handlers_list += [
            (r'^/api/webhooks/?$', handlers.WebhooksHandler)
        ]

    if settings.core.listen_support:
        handlers_list += [
            (r'^/api/listen/?$', handlers.ListenHandler)
        ]

    # reverse API calls

    if settings.reverse.enabled:
        handlers_list += [
            (r'^/api/reverse/?$', handlers.ReverseHandler)
        ]

    handlers_list += [
        (r'^.*$', handlers.NoSuchFunctionHandler)
    ]

    return handlers_list


def get_application():
    global _application

    if _application is None:
        _application = Application(
            handlers=_make_handlers(),
            debug=False,
            compress_response=settings.server.compress_response,
            log_function=_log_request,
        )

    return _application




from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd, run_set_cmd


class NetError(Exception):
    pass


def has_network_ip_support():
    return bool(settings.system.net.ip.get_cmd and settings.system.net.ip.set_cmd)


def get_ip_config():
    return run_get_cmd(settings.system.net.ip.get_cmd, cmd_name='IP config',
                       exc_class=NetError, required_fields=['ip', 'mask', 'gw', 'dns'])


def set_ip_config(ip, mask, gw, dns):
    run_set_cmd(settings.system.net.ip.set_cmd, cmd_name='IP config', exc_class=NetError,
                ip=ip, mask=mask, gw=gw, dns=dns)


def has_network_wifi_support():
    return bool(settings.system.net.wifi.get_cmd and settings.system.net.wifi.set_cmd)


def get_wifi_config():
    wifi_config = run_get_cmd(settings.system.net.wifi.get_cmd, cmd_name='WiFi config', log_values=False,
                              exc_class=NetError, required_fields=['ssid', 'psk', 'bssid'])

    return wifi_config


def set_wifi_config(ssid, psk, bssid):
    run_set_cmd(settings.system.net.wifi.set_cmd, cmd_name='WiFi config', exc_class=NetError, log_values=False,
                ssid=ssid, psk=psk, bssid=bssid)




import logging
import queue

from tornado.httpclient import AsyncHTTPClient, HTTPRequest

from qtoggleserver import persist
from qtoggleserver.conf import settings
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core import responses as core_responses
from qtoggleserver.utils import http as http_utils
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)
_webhooks = None


class InvalidParamError(Exception):
    def __init__(self, param):
        self._param = param
        super().__init__('invalid field: {}'.format(param))


class WebhooksRequest:
    def __init__(self, body):
        self.body = body
        self.retries = 0


class Webhooks:
    # noinspection PyUnusedLocal
    def __init__(self, scheme=None, host=None, port=None, path=None, timeout=None, retries=None, **kwargs):

        # the enabled value comes with kwargs but is ignored,
        # as the webhook will be explicitly enabled afterwards

        self._enabled = False
        self._scheme = scheme
        self._host = host
        self._port = port
        self._path = path
        self._timeout = timeout
        self._retries = retries
        self._url = None

        self._queue = queue.Queue(settings.core.event_queue_size)

    def __str__(self):
        s = 'webhooks'

        if not self._enabled:
            s += ' [disabled]'

        if self._scheme:
            s += ' ' + self.get_url()

        return s

    def is_enabled(self):
        return self._enabled

    def enable(self):
        if self._enabled:
            return

        logger.debug('enabling %s', self)

        self._enabled = True

    def disable(self):
        if not self._enabled:
            return

        logger.debug('disabling %s', self)

        self._enabled = False

        # drop all queued requests
        while not self._queue.empty():
            self._queue.get()

    def get_url(self):
        if not hasattr(self, '_url'):
            if self._scheme == 'http' and self._port == 80 or self._scheme == 'https' and self._port == 443:
                self._url = '{}://{}{}'.format(self._scheme, self._host, self._path)

            else:
                self._url = '{}://{}:{}{}'.format(self._scheme, self._host, self._port, self._path)

        return self._url

    def call(self, body):
        if not self._enabled:
            return

        request = WebhooksRequest(body)

        if self._queue.empty():
            return self._request(request)

        try:
            self._queue.put(request)

        except queue.Full:
            logger.error('%s: queue is full', self)
            return

    def _request(self, request):
        if not self._enabled:
            return

        url = self.get_url()
        body = json_utils.dumps(request.body)

        def on_response(response):
            try:
                core_responses.parse(response)

            except core_responses.Error as e:
                logger.error('%s: call failed: %s', self, e)

                # retry mechanism
                if not self._retries:
                    return self._check_pending()

                if request.retries <= self._retries:
                    request.retries += 1
                    logger.debug('%s: resending request (retry %s/%s)', self, request.retries, self._retries)

                    self._request(request)

                else:
                    self._check_pending()

            else:
                logger.debug('%s: call succeeded', self)
                self._check_pending()

        http_client = AsyncHTTPClient()
        # TODO use webhooks password
        headers = {
            'Content-Type': http_utils.JSON_CONTENT_TYPE,
            'Authorization': core_api_auth.make_auth_header(core_api_auth.ORIGIN_DEVICE, username=None,
                                                            password_hash=core_device_attrs.normal_password_hash)
        }
        request = HTTPRequest(url, 'POST', headers=headers, body=body,
                              connect_timeout=self._timeout, request_timeout=self._timeout,
                              follow_redirects=True)

        logger.debug('%s: calling', self)

        http_client.fetch(request, on_response)  # TODO await

    def _check_pending(self):
        try:
            request = self._queue.get(block=False)

        except queue.Empty:
            return

        self._request(request)

    def to_json(self):
        d = {
            'enabled': self._enabled,
            'scheme': self._scheme,
            'host': self._host,
            'port': self._port,
            'path': self._path,
            'timeout': self._timeout,
            'retries': self._retries
        }

        return d


def get():
    return _webhooks


def setup(enabled, scheme=None, host=None, port=None, path=None, timeout=None, retries=None, **kwargs):
    global _webhooks

    if _webhooks and _webhooks.is_enabled():
        _webhooks.disable()

    _webhooks = Webhooks(scheme, host, port, path, timeout, retries)
    if enabled:
        _webhooks.enable()


def load():
    data = persist.get_value('webhooks')
    if data is None:
        setup(enabled=False)
        logger.debug('loaded %s', _webhooks)

        return

    setup(**data)
    logger.debug('loaded %s', _webhooks)


def save():
    if _webhooks is None:
        return

    logger.debug('saving data')
    persist.set_value('webhooks', _webhooks.to_json())


def reset():
    logger.debug('clearing persisted data')
    persist.remove('webhooks')
